"""General tools for Docs."""

from lightning_utilities.docs.formatting import adjust_linked_external_docs
from lightning_utilities.docs.retriever import fetch_external_assets

__all__ = ["adjust_linked_external_docs", "fetch_external_assets"]



# Copyright The Lightning AI team.
# Licensed under the Apache License, Version 2.0 (the "License");
#     http://www.apache.org/licenses/LICENSE-2.0

import functools
import importlib
import os
import warnings
from functools import lru_cache
from importlib.metadata import PackageNotFoundError, distribution
from importlib.metadata import version as _version
from importlib.util import find_spec
from types import ModuleType
from typing import Any, Callable, Optional, TypeVar

from packaging.requirements import Requirement
from packaging.version import InvalidVersion, Version
from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")

try:
    from importlib import metadata
except ImportError:
    # Python < 3.8
    import importlib_metadata as metadata  # type: ignore


@lru_cache
def package_available(package_name: str) -> bool:
    """Check if a package is available in your environment.

    >>> package_available('os')
    True
    >>> package_available('bla')
    False

    """
    try:
        return find_spec(package_name) is not None
    except ModuleNotFoundError:
        return False


@lru_cache
def module_available(module_path: str) -> bool:
    """Check if a module path is available in your environment.

    >>> module_available('os')
    True
    >>> module_available('os.bla')
    False
    >>> module_available('bla.bla')
    False

    """
    module_names = module_path.split(".")
    if not package_available(module_names[0]):
        return False
    try:
        importlib.import_module(module_path)
    except ImportError:
        return False
    return True


def compare_version(package: str, op: Callable, version: str, use_base_version: bool = False) -> bool:
    """Compare package version with some requirements.

    >>> compare_version("torch", operator.ge, "0.1")
    True
    >>> compare_version("does_not_exist", operator.ge, "0.0")
    False

    """
    try:
        pkg = importlib.import_module(package)
    except (ImportError, RuntimeError):
        return False
    try:
        # Use importlib.metadata to infer version
        pkg_version = Version(pkg.__version__) if hasattr(pkg, "__version__") else Version(_version(package))
    except (TypeError, PackageNotFoundError):
        # this is mocked by Sphinx, so it should return True to generate all summaries
        return True
    if use_base_version:
        pkg_version = Version(pkg_version.base_version)
    return op(pkg_version, Version(version))


class RequirementCache:
    """Boolean-like class to check for requirement and module availability.

    Args:
        requirement: The requirement to check, version specifiers are allowed.
        module: The optional module to try to import if the requirement check fails.

    >>> RequirementCache("torch>=0.1")
    Requirement 'torch>=0.1' met
    >>> bool(RequirementCache("torch>=0.1"))
    True
    >>> bool(RequirementCache("torch>100.0"))
    False
    >>> RequirementCache("torch")
    Requirement 'torch' met
    >>> bool(RequirementCache("torch"))
    True
    >>> bool(RequirementCache("unknown_package"))
    False
    >>> bool(RequirementCache(module="torch.utils"))
    True
    >>> bool(RequirementCache(module="unknown_package"))
    False
    >>> bool(RequirementCache(module="unknown.module.path"))
    False

    """

    def __init__(self, requirement: Optional[str] = None, module: Optional[str] = None) -> None:
        if not (requirement or module):
            raise ValueError("At least one arguments need to be set.")
        self.requirement = requirement
        self.module = module

    def _check_requirement(self) -> None:
        if not self.requirement:
            raise ValueError("Requirement name is required.")
        try:
            req = Requirement(self.requirement)
            pkg_version = Version(_version(req.name))
            self.available = req.specifier.contains(pkg_version, prereleases=True) and (
                not req.extras or self._check_extras_available(req)
            )
        except (PackageNotFoundError, InvalidVersion) as ex:
            self.available = False
            self.message = f"{ex.__class__.__name__}: {ex}. HINT: Try running `pip install -U {self.requirement!r}`"

        if self.available:
            self.message = f"Requirement {self.requirement!r} met"
        else:
            req_include_version = any(c in self.requirement for c in "=<>")
            if not req_include_version or self.module is not None:
                module = self.requirement if self.module is None else self.module
                # Sometimes `importlib.metadata.version` fails but the module is importable
                self.available = module_available(module)
                if self.available:
                    self.message = f"Module {module!r} available"
            self.message = (
                f"Requirement {self.requirement!r} not met. HINT: Try running `pip install -U {self.requirement!r}`"
            )

    def _check_module(self) -> None:
        if not self.module:
            raise ValueError("Module name is required.")
        self.available = module_available(self.module)
        if self.available:
            self.message = f"Module {self.module!r} available"
        else:
            self.message = f"Module not found: {self.module!r}. HINT: Try running `pip install -U {self.module}`"

    def _check_available(self) -> None:
        if hasattr(self, "available"):
            return
        if self.requirement:
            self._check_requirement()
        if getattr(self, "available", True) and self.module:
            self._check_module()

    def _check_extras_available(self, requirement: Requirement) -> bool:
        if not requirement.extras:
            return True

        extra_requirements = self._get_extra_requirements(requirement)

        if not extra_requirements:
            # The specified extra is not found in the package metadata
            return False

        # Verify each extra requirement is installed
        for extra_req in extra_requirements:
            try:
                extra_dist = distribution(extra_req.name)
                extra_installed_version = Version(extra_dist.version)
                if extra_req.specifier and not extra_req.specifier.contains(extra_installed_version, prereleases=True):
                    return False
            except importlib.metadata.PackageNotFoundError:
                return False

        return True

    def _get_extra_requirements(self, requirement: Requirement) -> list[Requirement]:
        dist = distribution(requirement.name)
        # Get the required dependencies for the specified extras
        extra_requirements = dist.metadata.get_all("Requires-Dist") or []
        return [Requirement(r) for r in extra_requirements if any(extra in r for extra in requirement.extras)]

    def __bool__(self) -> bool:
        """Format as bool."""
        self._check_available()
        return self.available

    def __str__(self) -> str:
        """Format as string."""
        self._check_available()
        return self.message

    def __repr__(self) -> str:
        """Format as string."""
        return self.__str__()


class ModuleAvailableCache(RequirementCache):
    """Boolean-like class for check of module availability.

    >>> ModuleAvailableCache("torch")
    Module 'torch' available
    >>> bool(ModuleAvailableCache("torch.utils"))
    True
    >>> bool(ModuleAvailableCache("unknown_package"))
    False
    >>> bool(ModuleAvailableCache("unknown.module.path"))
    False

    """

    def __init__(self, module: str) -> None:
        warnings.warn(
            "`ModuleAvailableCache` is a special case of `RequirementCache`."
            " Please use `RequirementCache(module=...)` instead.",
            DeprecationWarning,
            stacklevel=4,
        )
        super().__init__(module=module)


def get_dependency_min_version_spec(package_name: str, dependency_name: str) -> str:
    """Return the minimum version specifier of a dependency of a package.

    >>> get_dependency_min_version_spec("pytorch-lightning==1.8.0", "jsonargparse")
    '>=4.12.0'

    """
    dependencies = metadata.requires(package_name) or []
    for dep in dependencies:
        dependency = Requirement(dep)
        if dependency.name == dependency_name:
            spec = [str(s) for s in dependency.specifier if str(s)[0] == ">"]
            return spec[0] if spec else ""
    raise ValueError(
        "This is an internal error. Please file a GitHub issue with the error message. Dependency "
        f"{dependency_name!r} not found in package {package_name!r}."
    )


class LazyModule(ModuleType):
    """Proxy module that lazily imports the underlying module the first time it is actually used.

    Args:
        module_name: the fully-qualified module name to import
        callback: a callback function to call before importing the module

    """

    def __init__(self, module_name: str, callback: Optional[Callable] = None) -> None:
        super().__init__(module_name)
        self._module: Any = None
        self._callback = callback

    def __getattr__(self, item: str) -> Any:
        """Lazily import the underlying module and delegate attribute access to it."""
        if self._module is None:
            self._import_module()

        return getattr(self._module, item)

    def __dir__(self) -> list[str]:
        """Lazily import the underlying module and return its attributes for introspection (dir())."""
        if self._module is None:
            self._import_module()

        return dir(self._module)

    def _import_module(self) -> None:
        # Execute callback, if any
        if self._callback is not None:
            self._callback()

        # Actually import the module
        self._module = importlib.import_module(self.__name__)

        # Update this object's dict so that attribute references are efficient
        # (__getattr__ is only called on lookups that fail)
        self.__dict__.update(self._module.__dict__)


def lazy_import(module_name: str, callback: Optional[Callable] = None) -> LazyModule:
    """Return a proxy module object that will lazily import the given module the first time it is used.

    Example usage:

        # Lazy version of `import tensorflow as tf`
        tf = lazy_import("tensorflow")
        # Other commands
        # Now the module is loaded
        tf.__version__

    Args:
        module_name: the fully-qualified module name to import
        callback: a callback function to call before importing the module

    Returns:
        a proxy module object that will be lazily imported when first used

    """
    return LazyModule(module_name, callback=callback)


def requires(*module_path_version: str, raise_exception: bool = True) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to check optional dependencies at call time with a clear error/warning message.

    Args:
        module_path_version: Python module paths (e.g., ``"torch.cuda"``) and/or pip-style requirements
            (e.g., ``"torch>=2.0.0"``) to verify.
        raise_exception: If ``True``, raise ``ModuleNotFoundError`` when requirements are not satisfied;
            otherwise emit a warning and proceed to call the function.

    Example:
        >>> @requires("libpath", raise_exception=bool(int(os.getenv("LIGHTING_TESTING", "0"))))
        ... def my_cwd():
        ...     from pathlib import Path
        ...     return Path(__file__).parent

        >>> class MyRndPower:
        ...     @requires("math", "random")
        ...     def __init__(self):
        ...         from math import pow
        ...         from random import randint
        ...         self._rnd = pow(randint(1, 9), 2)

    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        reqs = [
            ModuleAvailableCache(mod_ver) if "." in mod_ver else RequirementCache(mod_ver)
            for mod_ver in module_path_version
        ]
        available = all(map(bool, reqs))

        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            if not available:
                missing = os.linesep.join([repr(r) for r in reqs if not bool(r)])
                msg = f"Required dependencies not available: \n{missing}"
                if raise_exception:
                    raise ModuleNotFoundError(msg)
                warnings.warn(msg, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import inspect
import json
import logging
import time
import typing
import uuid
import warnings
from collections import deque
from collections.abc import AsyncGenerator, Iterator
from enum import Enum
from typing import Annotated, Literal, Optional, Union

from fastapi import BackgroundTasks, HTTPException, Request, Response
from fastapi.responses import StreamingResponse
from pydantic import BaseModel, Field

from litserve.constants import _DEFAULT_LIT_API_PATH
from litserve.specs.base import LitSpec, _AsyncSpecWrapper
from litserve.utils import LitAPIStatus, azip

if typing.TYPE_CHECKING:
    from litserve import LitAPI, LitServer

logger = logging.getLogger(__name__)


def shortuuid():
    return uuid.uuid4().hex[:6]


class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0
    completion_tokens: Optional[int] = 0

    def __add__(self, other: "UsageInfo") -> "UsageInfo":
        other.prompt_tokens += self.prompt_tokens
        other.completion_tokens += self.completion_tokens
        other.total_tokens += self.total_tokens
        return other

    def __radd__(self, other):
        if other == 0:
            return self
        return self.__add__(other)


class TextContent(BaseModel):
    type: str
    text: str


class ImageDetail(str, Enum):
    auto: str = "auto"
    low: str = "low"
    high: str = "high"


class ImageContentURL(BaseModel):
    url: str
    detail: ImageDetail = ImageDetail.auto


class ImageContent(BaseModel):
    type: str
    image_url: Union[str, ImageContentURL]


class InputAudio(BaseModel):
    data: str  # base64 encoded audio data.
    format: Literal["wav", "mp3"]


class AudioContent(BaseModel):
    type: Literal["input_audio"]
    input_audio: InputAudio


class Function(BaseModel):
    name: str
    description: str
    parameters: dict[str, object]


class ToolChoice(str, Enum):
    auto: str = "auto"
    none: str = "none"
    any: str = "any"


class Tool(BaseModel):
    type: Literal["function"]
    function: Function


class FunctionCall(BaseModel):
    name: str
    arguments: str


class ToolCall(BaseModel):
    index: int = 0
    id: Optional[str] = None
    type: str = "function"
    function: FunctionCall


class ResponseFormatText(BaseModel):
    type: Literal["text"]


class ResponseFormatJSONObject(BaseModel):
    type: Literal["json_object"]


class JSONSchema(BaseModel):
    name: str
    description: Optional[str] = None
    schema_: Optional[dict[str, object]] = Field(None, alias="schema")
    strict: Optional[bool] = False


class ResponseFormatJSONSchema(BaseModel):
    json_schema: JSONSchema
    type: Literal["json_schema"]


ResponseFormat = Annotated[
    Union[ResponseFormatText, ResponseFormatJSONObject, ResponseFormatJSONSchema], "ResponseFormat"
]


class ChatMessage(BaseModel):
    role: str
    content: Optional[Union[str, list[Union[TextContent, ImageContent, AudioContent]]]] = None
    name: Optional[str] = None
    tool_calls: Optional[list[ToolCall]] = None
    tool_call_id: Optional[str] = None


class ChatMessageWithUsage(ChatMessage):
    prompt_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0
    completion_tokens: Optional[int] = 0


class ChoiceDelta(ChatMessage):
    content: Optional[str] = None
    role: Optional[Literal["system", "user", "assistant", "tool"]] = None


class ChatCompletionRequest(BaseModel):
    model: Optional[str] = ""
    messages: list[ChatMessage]
    temperature: Optional[float] = 0.7
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    max_tokens: Optional[int] = None  # Kept for backward compatibility
    max_completion_tokens: Optional[int] = None
    stop: Optional[Union[str, list[str]]] = None
    stream: Optional[bool] = False
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    user: Optional[str] = None
    tools: Optional[list[Tool]] = None
    tool_choice: Optional[ToolChoice] = ToolChoice.auto
    response_format: Optional[ResponseFormat] = None
    reasoning_effort: Optional[Literal["low", "medium", "high"]] = None
    metadata: Optional[dict[str, str]] = None


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatMessage
    finish_reason: Optional[Literal["stop", "length"]]


class ChatCompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"chatcmpl-{shortuuid()}")
    object: str = "chat.completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: list[ChatCompletionResponseChoice]
    usage: UsageInfo


class ChatCompletionStreamingChoice(BaseModel):
    delta: Optional[ChoiceDelta]
    finish_reason: Optional[Literal["stop", "length", "tool_calls", "content_filter", "function_call"]] = None
    index: int
    logprobs: Optional[dict] = None


class ChatCompletionChunk(BaseModel):
    id: str = Field(default_factory=lambda: f"chatcmpl-{shortuuid()}")
    object: str = "chat.completion.chunk"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    system_fingerprint: Optional[str] = None
    choices: list[ChatCompletionStreamingChoice]
    usage: Optional[UsageInfo]


LITAPI_VALIDATION_MSG = """LitAPI.predict and LitAPI.encode_response must be a generator (use yield instead or return)
while using the OpenAISpec.

Error: {}

Please follow the below examples for guidance on how to use the spec:

If your current code looks like this:

```
import litserve as ls
from litserve.specs.openai import ChatMessage

class ExampleAPI(ls.LitAPI):
    ...
    def predict(self, x):
        return "This is a generated output"

    def encode_response(self, output: dict):
        return ChatMessage(role="assistant", content="This is a custom encoded output")
```

You should modify it to:

```
import litserve as ls
from litserve.specs.openai import ChatMessage

class ExampleAPI(ls.LitAPI):
    ...
    def predict(self, x):
        yield "This is a generated output"

    def encode_response(self, output):
        yield ChatMessage(role="assistant", content="This is a custom encoded output")
```


You can also yield responses in chunks. LitServe will handle the streaming for you:

```
class ExampleAPI(ls.LitAPI):
    ...
    def predict(self, x):
        yield from self.model(x)

    def encode_response(self, output):
        for out in output:
            yield ChatMessage(role="assistant", content=out)
```
"""

ASYNC_LITAPI_VALIDATION_MSG = """Error: {}

`enable_async` is set but LitAPI method is not async. To use async with OpenAISpec, you need to make the following changes:

- LitAPI.decode_request can be a regular function or an async function.
- LitAPI.predict must be an async generator (use 'yield' or 'yield from' inside an 'async def' function).
- LitAPI.encode_response can be a regular function or an async generator.


Please follow the examples below for guidance on how to use the spec in async mode:

If your current code looks like this:

```
import litserve as ls
from litserve.specs.openai import ChatMessage

class ExampleAPI(ls.LitAPI):
    ...
    def predict(self, x):
        return "This is a generated output"

    def encode_response(self, output: dict):
        return ChatMessage(role="assistant", content="This is a custom encoded output")
```
You should modify it to:

```
import litserve as ls
from litserve.specs.openai import ChatMessage

class ExampleAPI(ls.LitAPI):
    ...
    async def decode_request(self, request):
        # process request here
        return request.messages

    async def predict(self, x):
        yield "This is a generated output"

    async def encode_response(self, output):
        yield ChatMessage(role="assistant", content="This is a custom encoded output")
```
You can also yield responses in chunks. LitServe will handle the streaming for you:

```
class ExampleAPI(ls.LitAPI):
    ...
    async def decode_request(self, request):
        # process request here
        return request.messages

    async def predict(self, x):
        async for out in self.model(x):
            yield out

    async def encode_response(self, output):
        async for out in output:
            yield ChatMessage(role="assistant", content=out)
```
"""  # noqa: E501


def _openai_format_error(error: Exception):
    if isinstance(error, HTTPException):
        return "data: " + json.dumps(
            {
                "error": {
                    "message": error.detail,
                    "type": "internal",
                    "param": None,
                    "code": "internal_error",
                }
            }
        )
    return "data: " + json.dumps(
        {
            "error": {
                "message": "Internal server error",
                "type": "internal",
                "param": None,
                "code": "internal_error",
            }
        }
    )


class OpenAISpec(LitSpec):
    def __init__(
        self,
    ):
        super().__init__()
        self.api_path = "/v1/chat/completions"  # default api path

    @property
    def stream(self):
        return True

    def pre_setup(self, lit_api: "LitAPI"):
        from litserve import LitAPI

        # Override the spec's api_path only if provided
        if lit_api._api_path and lit_api._api_path not in (_DEFAULT_LIT_API_PATH, self.api_path):
            self.api_path = lit_api._api_path
            warnings.warn(
                f"Custom API path detected: '{self.api_path}'. "
                "The OpenAI SDK only supports the default path '/v1/chat/completions'. "
                f"To use '{self.api_path}', send HTTP requests directly or use a client that supports custom endpoints."
                " For SDK compatibility, use the default path."
            )

        # register the endpoint
        self.add_endpoint(self.api_path, self.chat_completion, ["POST"])
        self.add_endpoint(self.api_path, self.options_chat_completions, ["OPTIONS"])

        # validate LitAPI methods
        is_encode_response_original = lit_api.encode_response.__code__ is LitAPI.encode_response.__code__

        if lit_api.enable_async:
            # warning for decode_request and encode_response
            if not asyncio.iscoroutinefunction(lit_api.decode_request):
                logger.info("decode_request is not a coroutine function. LitServe will asyncify it.")
            if not inspect.isasyncgenfunction(lit_api.encode_response):
                logger.info("encode_response is not an async generator. LitServe will asyncify it.")

            if not inspect.isasyncgenfunction(lit_api.predict):
                raise ValueError(ASYNC_LITAPI_VALIDATION_MSG.format("predict must be an async generator"))

            if (
                not is_encode_response_original
                and not inspect.isgeneratorfunction(lit_api.encode_response)
                and not inspect.isasyncgenfunction(lit_api.encode_response)
            ):
                raise ValueError(
                    ASYNC_LITAPI_VALIDATION_MSG.format("encode_response is neither a generator nor an async generator")
                )

        else:
            for method in ["decode_request", "predict", "encode_response"]:
                method_obj = getattr(lit_api, method)
                if asyncio.iscoroutinefunction(method_obj) or inspect.isasyncgenfunction(method_obj):
                    raise ValueError(
                        f"'{method}' is defined as async/coroutine, but 'enable_async' is not set in LitAPI. "
                        "Please set 'enable_async=True' in your LitAPI class or "
                        "use synchronous (non-async) methods instead."
                    )

            if not inspect.isgeneratorfunction(lit_api.predict):
                raise ValueError(LITAPI_VALIDATION_MSG.format("predict is not a generator"))

            if not is_encode_response_original and not inspect.isgeneratorfunction(lit_api.encode_response):
                raise ValueError(LITAPI_VALIDATION_MSG.format("encode_response is not a generator"))

    def setup(self, server: "LitServer"):
        super().setup(server)
        print("OpenAI spec setup complete")

    def as_async(self) -> "_AsyncOpenAISpecWrapper":
        return _AsyncOpenAISpecWrapper(self)

    def populate_context(self, context, request):
        data = request.dict()
        data.pop("messages")
        context.update(data)

    def decode_request(
        self, request: ChatCompletionRequest, context_kwargs: Optional[dict] = None
    ) -> ChatCompletionRequest:
        return request

    def batch(self, inputs):
        return list(inputs)

    def unbatch(self, output):
        yield output

    def extract_usage_info(self, output: dict) -> dict:
        prompt_tokens: int = output.pop("prompt_tokens", 0)
        completion_tokens: int = output.pop("completion_tokens", 0)
        total_tokens: int = output.pop("total_tokens", 0)
        return {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": total_tokens,
        }

    def validate_chat_message(self, obj):
        return isinstance(obj, dict) and "role" in obj and "content" in obj

    def _encode_response(self, output: Union[dict[str, str], list[dict[str, str]]]) -> dict:
        logger.debug(output)
        if output is None:
            message = {"role": "assistant", "content": None}
        elif isinstance(output, str):
            message = {"role": "assistant", "content": output}
        elif self.validate_chat_message(output):
            message = output
        elif isinstance(output, dict) and "content" in output:
            message = output.copy()
            message.update(role="assistant")
        elif isinstance(output, list) and output and self.validate_chat_message(output[-1]):
            message = output[-1]
        else:
            error = (
                "Malformed output from LitAPI.predict: expected "
                f"string or {{'role': '...', 'content': '...'}}, got '{output}'."
            )
            logger.error(error)
            raise HTTPException(500, error)
        usage_info = self.extract_usage_info(message)
        return {**message, **usage_info}

    def encode_response(
        self, output_generator: Union[dict[str, str], list[dict[str, str]]], context_kwargs: Optional[dict] = None
    ) -> Iterator[Union[ChatMessage, ChatMessageWithUsage]]:
        for output in output_generator:
            logger.debug(output)
            yield self._encode_response(output)

    async def get_from_queues(self, uids) -> list[AsyncGenerator]:
        choice_pipes = []
        for uid, q, event in zip(uids, self.queues, self.events):
            data = self.data_streamer(q, event, send_status=True)
            choice_pipes.append(data)
        return choice_pipes

    async def options_chat_completions(self, request: Request):
        return Response(status_code=200)

    async def chat_completion(self, request: ChatCompletionRequest, background_tasks: BackgroundTasks):
        response_queue_id = self.response_queue_id
        logger.debug("Received chat completion request %s", request)
        uids = [uuid.uuid4() for _ in range(request.n)]
        self.queues = []
        self.events = []
        for uid in uids:
            request_el = request.model_copy()
            request_el.n = 1
            q = deque()
            event = asyncio.Event()
            self.response_buffer[uid] = (q, event)
            self.request_queue.put((response_queue_id, uid, time.monotonic(), request_el))
            self.queues.append(q)
            self.events.append(event)

        responses = await self.get_from_queues(uids)

        if request.stream:
            return StreamingResponse(
                self.streaming_completion(request, responses),
                media_type="text/event-stream",
                background=background_tasks,
            )

        response_task = asyncio.create_task(self.non_streaming_completion(request, responses))
        return await response_task

    async def streaming_completion(self, request: ChatCompletionRequest, pipe_responses: list):
        try:
            model = request.model
            usage_info = None
            async for streaming_response in azip(*pipe_responses):
                choices = []
                usage_infos = []
                # iterate over n choices
                for i, (response, status) in enumerate(streaming_response):
                    if status == LitAPIStatus.ERROR and isinstance(response, HTTPException):
                        raise response
                    elif status == LitAPIStatus.ERROR:
                        logger.error("Error in streaming response: %s", response)
                        raise HTTPException(status_code=500)
                    encoded_response = json.loads(response)
                    logger.debug(encoded_response)
                    chat_msg = ChoiceDelta(**encoded_response)
                    usage_infos.append(UsageInfo(**encoded_response))
                    choice = ChatCompletionStreamingChoice(
                        index=i, delta=chat_msg, system_fingerprint="", finish_reason=None
                    )

                    choices.append(choice)

                # Only use the last item from encode_response
                usage_info = sum(usage_infos)
                chunk = ChatCompletionChunk(model=model, choices=choices, usage=None)
                logger.debug(chunk)
                yield f"data: {chunk.model_dump_json(by_alias=True)}\n\n"

            choices = [
                ChatCompletionStreamingChoice(
                    index=i,
                    delta=ChoiceDelta(),
                    finish_reason="stop",
                )
                for i in range(request.n)
            ]
            last_chunk = ChatCompletionChunk(
                model=model,
                choices=choices,
                usage=usage_info,
            )
            yield f"data: {last_chunk.model_dump_json(by_alias=True)}\n\n"
            yield "data: [DONE]\n\n"
        except Exception as e:
            logger.error("Error in streaming response: %s", e, exc_info=True)
            yield _openai_format_error(e)
            return

    async def non_streaming_completion(self, request: ChatCompletionRequest, generator_list: list[AsyncGenerator]):
        try:
            model = request.model
            usage_infos = []
            choices = []
            # iterate over n choices
            for i, streaming_response in enumerate(generator_list):
                msgs = []
                tool_calls = None
                usage = None
                async for response, status in streaming_response:
                    if status == LitAPIStatus.ERROR and isinstance(response, HTTPException):
                        raise response
                    if status == LitAPIStatus.ERROR:
                        logger.error("Error in OpenAI non-streaming response: %s", response)
                        raise HTTPException(status_code=500)

                    # data from LitAPI.encode_response
                    encoded_response = json.loads(response)
                    logger.debug(encoded_response)
                    chat_msg = ChatMessage(**encoded_response)
                    usage = UsageInfo(**encoded_response)
                    usage_infos.append(usage)  # Aggregate usage info across all choices
                    msgs.append(chat_msg.content)
                    if chat_msg.tool_calls:
                        tool_calls = chat_msg.tool_calls

                content = "".join(msg for msg in msgs if msg is not None)
                msg = {"role": "assistant", "content": content, "tool_calls": tool_calls}
                choice = ChatCompletionResponseChoice(index=i, message=msg, finish_reason="stop")
                choices.append(choice)

            return ChatCompletionResponse(model=model, choices=choices, usage=sum(usage_infos))
        except HTTPException as e:
            raise e
        except Exception as e:
            logger.error("Error in non-streaming response: %s", e, exc_info=True)
            raise HTTPException(status_code=500)


class _AsyncOpenAISpecWrapper(_AsyncSpecWrapper):
    async def encode_response(self, output_generator: AsyncGenerator, context_kwargs: Optional[dict] = None):
        async for output in output_generator:
            yield self._spec._encode_response(output)



"""Consume 10K tokens from the stream endpoint and measure the speed."""

import logging
import time

import requests
from requests.adapters import HTTPAdapter
from tenacity import retry, stop_after_attempt
from urllib3.util import Retry

logger = logging.getLogger(__name__)
# Configuration
SERVER_URL = "http://0.0.0.0:8000/predict"
TOTAL_TOKENS = 10000
EXPECTED_TTFT = 0.005  # time to first token

# tokens per second
MAX_SPEED = 3600  # 3600 on GitHub CI, 10000 on M3 Pro


def create_session(pool_connections=10, pool_maxsize=10, max_retries=3):
    """Create a session object with custom connection pool settings."""
    session = requests.Session()
    retry_strategy = Retry(
        total=max_retries,
        backoff_factor=0.1,
        # Don't retry on streaming requests
        allowed_methods=frozenset(["GET", "POST"]),
    )
    adapter = HTTPAdapter(pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=retry_strategy)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


# Initialize session with reasonable defaults for streaming
session = create_session(pool_connections=10, pool_maxsize=10)


def speed_test():
    start = time.time()
    resp = session.post(SERVER_URL, stream=True, json={"input": 1})
    num_tokens = 0
    ttft = None  # time to first token
    for line in resp.iter_lines():
        if not line:
            continue
        if ttft is None:
            ttft = time.time() - start
            print(f"Time to first token: {ttft}")
            assert ttft < EXPECTED_TTFT, f"Expected time to first token to be less than 0.1 seconds but got {ttft}"
        num_tokens += 1
    end = time.time()
    resp.raise_for_status()
    assert num_tokens == TOTAL_TOKENS, f"Expected {TOTAL_TOKENS} tokens, got {num_tokens}"
    speed = num_tokens / (end - start)
    return {"speed": speed, "time": end - start}


@retry(stop=stop_after_attempt(10))
def main():
    for i in range(10):
        try:
            resp = session.get("http://localhost:8000/health")
            if resp.status_code == 200:
                break
        except requests.exceptions.ConnectionError as e:
            logger.error(f"Error connecting to server: {e}")
        time.sleep(10)
    data = speed_test()
    speed = data["speed"]
    print(data)
    assert speed >= MAX_SPEED, f"Expected streaming speed to be greater than {MAX_SPEED}, got {speed}"


if __name__ == "__main__":
    main()



import asyncio
import pickle
from queue import Empty
from unittest.mock import AsyncMock, Mock, patch

import pytest
import zmq

from litserve.transport.zmq_queue import AsyncConsumer, Broker, Producer


@pytest.fixture
def mock_context():
    with patch("zmq.Context") as mock_ctx:
        socket = Mock()
        mock_ctx.return_value.socket.return_value = socket
        yield mock_ctx, socket


@pytest.fixture
def mock_async_context():
    with patch("zmq.asyncio.Context") as mock_ctx:
        socket = AsyncMock()  # Use AsyncMock for async methods
        mock_ctx.return_value.socket.return_value = socket
        yield mock_ctx, socket


def test_broker_start_stop(mock_context):
    _, socket = mock_context
    broker = Broker(use_process=False)

    # Start broker
    broker.start()
    assert socket.bind.call_count == 2  # Should bind both frontend and backend

    # Stop broker
    broker.stop()
    assert socket.close.call_count == 2  # Should close both sockets


def test_broker_error_handling(mock_context):
    """Test broker handles ZMQ errors."""
    _, socket = mock_context
    socket.bind.side_effect = zmq.ZMQError("Test error")

    broker = Broker(use_process=False)
    broker.start()
    broker.stop()

    assert socket.close.called  # Should clean up even on error


def test_producer_send(mock_context):
    _, socket = mock_context
    producer = Producer(address="test_addr")

    # Test sending simple data
    producer.put("test_data", consumer_id=1)
    sent_message = socket.send.call_args[0][0]
    consumer_id, data = sent_message.split(b"|", 1)
    assert consumer_id == b"1"
    assert pickle.loads(data) == "test_data"

    # Test sending complex data
    complex_data = {"key": [1, 2, 3]}
    producer.put(complex_data, consumer_id=2)
    sent_message = socket.send.call_args[0][0]
    consumer_id, data = sent_message.split(b"|", 1)
    assert consumer_id == b"2"
    assert pickle.loads(data) == complex_data


def test_producer_error_handling(mock_context):
    _, socket = mock_context
    producer = Producer(address="test_addr")

    # Test ZMQ error
    socket.send.side_effect = zmq.ZMQError("Test error")
    with pytest.raises(zmq.ZMQError):
        producer.put("data", consumer_id=1)

    # Test unpickleable object
    class Unpickleable:
        def __reduce__(self):
            raise pickle.PickleError("Can't pickle this!")

    with pytest.raises(pickle.PickleError):
        producer.put(Unpickleable(), consumer_id=1)


def test_producer_wait_for_subscribers(mock_context):
    _, socket = mock_context
    producer = Producer(address="test_addr")

    # Test successful wait
    assert producer.wait_for_subscribers(timeout=0.1)
    assert socket.send.called

    # Test timeout
    socket.send.side_effect = zmq.ZMQError("Would block")
    assert not producer.wait_for_subscribers(timeout=0.1)


@pytest.mark.parametrize("timeout", [1.0, None])
@pytest.mark.asyncio
async def test_async_consumer(mock_async_context, timeout):
    _, socket = mock_async_context
    consumer = AsyncConsumer(consumer_id=1, address="test_addr")

    # Setup mock received data
    test_data = {"test": "data"}
    message = b"1|" + pickle.dumps(test_data)
    socket.recv.return_value = message

    # Test receiving
    received = await consumer.get(timeout=timeout)
    assert received == test_data

    # Test timeout
    socket.recv.side_effect = asyncio.TimeoutError()
    with pytest.raises(Empty):
        await consumer.get(timeout=timeout)


@pytest.mark.asyncio
async def test_async_consumer_cleanup():
    with patch("zmq.asyncio.Context") as mock_ctx:
        socket = AsyncMock()
        mock_ctx.return_value.socket.return_value = socket

        consumer = AsyncConsumer(consumer_id=1, address="test_addr")
        consumer.close()

        assert socket.close.called
        assert mock_ctx.return_value.term.called


def test_producer_cleanup():
    with patch("zmq.Context") as mock_ctx:
        socket = Mock()
        mock_ctx.return_value.socket.return_value = socket

        producer = Producer(address="test_addr")
        producer.close()

        assert socket.close.called
        assert mock_ctx.return_value.term.called



import base64
import io

import PIL
import torch
import torchvision
from fastapi import FastAPI, HTTPException
from jsonargparse import CLI
from pydantic import BaseModel

# Set float32 matrix multiplication precision if GPU is available and capable
if torch.cuda.is_available() and torch.cuda.get_device_capability() >= (8, 0):
    torch.set_float32_matmul_precision("high")

app = FastAPI()


class ImageData(BaseModel):
    image_data: str


class ImageClassifierAPI:
    def __init__(self, device):
        self.device = device
        weights = torchvision.models.ResNet18_Weights.DEFAULT
        self.image_processing = weights.transforms()
        self.model = torchvision.models.resnet18(weights=None).eval().to(device)

    def process_image(self, image_data):
        image = base64.b64decode(image_data)
        pil_image = PIL.Image.open(io.BytesIO(image)).convert("RGB")
        processed_image = self.image_processing(pil_image)
        return processed_image.unsqueeze(0).to(self.device)  # Add batch dimension

    def predict(self, x):
        with torch.inference_mode():
            outputs = self.model(x)
            _, predictions = torch.max(outputs, 1)
        return predictions.item()


device = "cuda" if torch.cuda.is_available() else "cpu"
if torch.backends.mps.is_available():
    device = "mps"
api = ImageClassifierAPI(device)


@app.get("/health")
async def health():
    return {"status": "ok"}


@app.post("/predict")
async def predict(image_data: ImageData):
    try:
        processed_image = api.process_image(image_data.image_data)
        prediction = api.predict(processed_image)
        return {"output": prediction}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


def main():
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8001, log_level="warning")


if __name__ == "__main__":
    CLI(main)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import litserve as ls
from litserve import docker_builder


def test_color():
    assert docker_builder.color("hi", docker_builder.RED) == f"{docker_builder.RED}hi{docker_builder.RESET}"

    expected = f"{docker_builder.INFO} {docker_builder.RED}hi{docker_builder.RESET}"
    assert docker_builder.color("hi", docker_builder.RED, docker_builder.INFO) == expected


EXPECTED_CONENT = f"""ARG PYTHON_VERSION=3.12
FROM python:$PYTHON_VERSION-slim

####### Add your own installation commands here #######
# RUN pip install some-package
# RUN wget https://path/to/some/data/or/weights
# RUN apt-get update && apt-get install -y <package-name>

WORKDIR /app
COPY . /app

# Install litserve and requirements
RUN pip install --no-cache-dir litserve=={ls.__version__} -r requirements.txt
EXPOSE 8000
CMD ["python", "/app/app.py"]
"""


EXPECTED_GPU_DOCKERFILE = f"""# Change CUDA and cuDNN version here
FROM nvidia/cuda:12.4.1-base-ubuntu22.04
ARG PYTHON_VERSION=3.12

ENV DEBIAN_FRONTEND=noninteractive
RUN apt-get update && apt-get install -y --no-install-recommends \\
        software-properties-common \\
        wget \\
    && add-apt-repository ppa:deadsnakes/ppa \\
    && apt-get update && apt-get install -y --no-install-recommends \\
        python$PYTHON_VERSION \\
        python$PYTHON_VERSION-dev \\
        python$PYTHON_VERSION-venv \\
    && wget https://bootstrap.pypa.io/get-pip.py -O get-pip.py \\
    && python$PYTHON_VERSION get-pip.py \\
    && rm get-pip.py \\
    && ln -sf /usr/bin/python$PYTHON_VERSION /usr/bin/python \\
    && ln -sf /usr/local/bin/pip$PYTHON_VERSION /usr/local/bin/pip \\
    && python --version \\
    && pip --version \\
    && apt-get purge -y --auto-remove software-properties-common \\
    && apt-get clean \\
    && rm -rf /var/lib/apt/lists/*

####### Add your own installation commands here #######
# RUN pip install some-package
# RUN wget https://path/to/some/data/or/weights
# RUN apt-get update && apt-get install -y <package-name>

WORKDIR /app
COPY . /app

# Install litserve and requirements
RUN pip install --no-cache-dir litserve=={ls.__version__} -r requirements.txt
EXPOSE 8000
CMD ["python", "/app/app.py"]
"""


def test_dockerize(tmp_path, monkeypatch):
    with open(tmp_path / "app.py", "w") as f:
        f.write("print('hello')")

    # Temporarily change the current working directory to tmp_path
    monkeypatch.chdir(tmp_path)

    with pytest.warns(UserWarning, match="Make sure to install the required packages in the Dockerfile."):
        docker_builder.dockerize("app.py", 8000)

    with open(tmp_path / "requirements.txt", "w") as f:
        f.write("lightning")

    docker_builder.dockerize("app.py", 8000)
    with open("Dockerfile") as f:
        content = f.read()
    assert content == EXPECTED_CONENT

    docker_builder.dockerize("app.py", 8000, gpu=True)
    with open("Dockerfile") as f:
        content = f.read()
    assert content == EXPECTED_GPU_DOCKERFILE

    with pytest.raises(FileNotFoundError, match="must be in the current directory"):
        docker_builder.dockerize("random_file_name.py", 8000)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import time

import numpy as np
import pytest
import torch
from fastapi import HTTPException
from pydantic import BaseModel

import litserve as ls
from litserve.specs.openai import ChatCompletionRequest


class TestDefaultBatchedAPI(ls.LitAPI):
    def setup(self, device) -> None:
        self.model = lambda x: len(x)

    def decode_request(self, request):
        return request["input"]

    def predict(self, x):
        return self.model(x)

    def encode_response(self, output):
        return {"output": output}


class TestCustomBatchedAPI(TestDefaultBatchedAPI):
    def batch(self, inputs):
        return np.stack(inputs)

    def unbatch(self, output):
        return list(output)


class TestStreamAPI(ls.LitAPI):
    def setup(self, device) -> None:
        self.model = None

    def decode_request(self, request):
        return request["input"]

    def predict(self, x):
        # x is a list of integers
        for i in range(4):
            yield np.asarray(x) * i

    def encode_response(self, output_stream):
        for output in output_stream:
            output = list(output)
            yield [{"output": o} for o in output]


def test_default_batch_unbatch():
    api = TestDefaultBatchedAPI(max_batch_size=4)
    api.request_timeout = 30
    api.pre_setup(spec=None)
    inputs = [1, 2, 3, 4]
    output = api.batch(inputs)
    assert output == inputs, "Default batch should not change input"
    assert api.unbatch(output) == inputs, "Default unbatch should not change input"


def test_default_unbatch_not_implemented():
    api = TestDefaultBatchedAPI()
    with pytest.raises(ValueError, match="Default implementation for `LitAPI.unbatch` method was not found."):
        api.unbatch(None)


class TestStreamAPIBatched(TestStreamAPI):
    def predict(self, x):
        for i in range(4):
            yield np.asarray(x) * i


def test_default_batch_unbatch_stream():
    api = TestStreamAPIBatched(max_batch_size=4)
    api.request_timeout = 30
    api.stream = True
    api.pre_setup(spec=None)
    inputs = [1, 2, 3, 4]
    expected_output = [[0, 0, 0, 0], [1, 2, 3, 4], [2, 4, 6, 8], [3, 6, 9, 12]]
    output = api.batch(inputs)
    output = api.predict(output)
    for out in api.unbatch(output):
        expected = expected_output.pop(0)
        assert np.all(out == expected), f"Default unbatch should not change input {out} != {expected}"


def test_custom_batch_unbatch():
    api = TestCustomBatchedAPI(max_batch_size=4)
    api.request_timeout = 30
    api.pre_setup(spec=None)
    inputs = [1, 2, 3, 4]
    output = api.batch(inputs)
    assert np.all(output == np.array(inputs)), "Custom batch stacks input as numpy array"
    assert api.unbatch(output) == inputs, "Custom unbatch should unstack input as list"


def test_batch_unbatch_stream():
    api = TestStreamAPI(max_batch_size=4)
    api.request_timeout = 30
    api.pre_setup(spec=None)
    inputs = [1, 2, 3, 4]
    output = api.batch(inputs)
    output = api.predict(output)
    output = api.unbatch(output)
    output = api.encode_response(output)
    first_resp = [o["output"] for o in next(output)]
    expected_outputs = [[0, 0, 0, 0], [1, 2, 3, 4], [2, 4, 6, 8], [3, 6, 9, 12]]
    assert first_resp == expected_outputs[0], "First response should be 0s"
    count = 1
    for out, expected_output in zip(output, expected_outputs[1:]):
        resp = [o["output"] for o in out]
        assert resp == expected_output
        count += 1

    assert count == 4, "Should have 4 responses"


def test_decode_request():
    request = {"input": 4.0}
    api = ls.test_examples.SimpleLitAPI()
    assert api.decode_request(request) == 4.0, "Decode request should return the input 4.0"


def test_decode_request_with_openai_spec():
    api = ls.test_examples.TestAPI(max_batch_size=1)
    api.request_timeout = 30
    api.pre_setup(spec=ls.OpenAISpec())
    request = ChatCompletionRequest(messages=[{"role": "system", "content": "Hello"}])
    decoded_request = api.decode_request(request)
    assert decoded_request.messages[0].content == "Hello", "Decode request should return the input message"


def test_encode_response():
    response = 4.0
    api = ls.test_examples.SimpleLitAPI()
    assert api.encode_response(response) == {"output": 4.0}, 'Encode response returns encoded output {"output": 4.0}'


def test_encode_response_with_openai_spec():
    api = ls.test_examples.TestAPI(max_batch_size=1)
    api.request_timeout = 30
    api.pre_setup(spec=ls.OpenAISpec())
    response = ["This", "is", "a", "LLM", "generated", "text"]
    generated_tokens = []
    for output in api.encode_response(response):
        generated_tokens.append(output["content"])
    assert generated_tokens == response, f"Encode response should return the generated tokens {response}"


def test_encode_response_with_openai_spec_dict_token_usage():
    prompt = "This is a LLM generated text"

    def predict():
        for token in prompt.split():
            yield {"content": token, "prompt_tokens": 4, "completion_tokens": 4, "total_tokens": 8}

    generated_tokens = []
    api = ls.test_examples.TestAPI(max_batch_size=1)
    api.request_timeout = 30
    api.pre_setup(spec=ls.OpenAISpec())

    for output in api.encode_response(predict()):
        assert output["role"] == "assistant", "Role should be assistant"
        generated_tokens.append(output["content"])
    assert generated_tokens == prompt.split(), f"Encode response should return the generated tokens {prompt.split()}"


def test_encode_response_with_custom_spec_api():
    class CustomSpecAPI(ls.OpenAISpec):
        def encode_response(self, output_stream):
            for output in output_stream:
                yield {"content": output}

    api = ls.test_examples.TestAPI(max_batch_size=1)
    api.request_timeout = 30
    api.pre_setup(spec=CustomSpecAPI())
    response = ["This", "is", "a", "LLM", "generated", "text"]
    generated_tokens = []
    for output in api.encode_response(response):
        generated_tokens.append(output["content"])
    assert generated_tokens == response, f"Encode response should return the generated tokens {response}"


def test_encode_response_with_openai_spec_invalid_input():
    api = ls.test_examples.TestAPI(max_batch_size=1)
    api.request_timeout = 30
    api.pre_setup(spec=ls.OpenAISpec())
    response = 10
    with pytest.raises(TypeError, match="object is not iterable"):
        next(api.encode_response(response))


def test_encode_response_with_openai_spec_invalid_predict_output():
    def predict():
        yield {"hello": "world"}

    api = ls.test_examples.TestAPI(max_batch_size=1)
    api.request_timeout = 30
    api.pre_setup(spec=ls.OpenAISpec())
    with pytest.raises(HTTPException, match=r"Malformed output from LitAPI.predict"):
        next(api.encode_response(predict()))


def test_format_encoded_response():
    api = ls.test_examples.SimpleLitAPI()
    sample = {"output": 4.0}
    msg = "Format encoded response should return the encoded response as a string"
    assert api.format_encoded_response(sample) == '{"output": 4.0}\n', msg

    class Sample(BaseModel):
        output: float
        name: str

    sample = Sample(output=4.0, name="test")
    msg = "Format encoded response should return the encoded response as a json string"
    assert json.loads(api.format_encoded_response(sample)) == {"output": 4.0, "name": "test"}, msg

    msg = "non dict and non Pydantic objects are returned as it is."
    assert api.format_encoded_response([1, 2, 3, 4]) == [1, 2, 3, 4], msg


def test_batch_torch():
    api = ls.test_examples.SimpleLitAPI()
    x = [torch.Tensor([1, 2, 3, 4]), torch.Tensor([5, 6, 7, 8])]
    assert torch.all(api.batch(x) == torch.stack(x)), "Batch should stack torch tensors"


def test_batch_numpy():
    api = ls.test_examples.SimpleLitAPI()
    x = [np.asarray([1, 2, 3, 4]), np.asarray([5, 6, 7, 8])]
    assert np.all(api.batch(x) == np.stack(x)), "Batch should stack Numpy array"


def test_device_property():
    api = ls.test_examples.SimpleLitAPI()
    api.device = "cpu"
    assert api.device == "cpu"


class TestLogger(ls.Logger):
    def process(self, key, value):
        self.processed_data = (key, value)


def test_log():
    api = ls.test_examples.SimpleLitAPI()
    api.request_timeout = 30
    assert api._logger_queue is None, "Logger queue should be None"
    assert api.log("time", 0.1) is None, "Log should return None"
    with pytest.warns(UserWarning, match="attempted without a configured logger"):
        api.log("time", 0.1)

    api = ls.test_examples.SimpleLitAPI()
    api.request_timeout = 30
    assert api._logger_queue is None, "Logger queue should be None"
    server = ls.LitServer(api, loggers=TestLogger())
    server._init_manager(1)
    server._logger_connector.run(server)
    server.launch_inference_worker(api)
    api.log("time", 0.1)
    assert server.logger_queue.get() == ("time", 0.1)


def test_enable_async_not_set():
    with pytest.raises(
        ValueError, match=r"predict must be an async generator or async function when enable_async=True"
    ):
        ls.test_examples.SimpleLitAPI(enable_async=True)


class HeavyInitAPI(ls.LitAPI):
    def __init__(self):
        super().__init__()
        time.sleep(2)


class HeavyInitAPIWithSetup(HeavyInitAPI):
    def setup(self, device):
        pass


class HeavySetupAPI(ls.LitAPI):
    def setup(self, device):
        time.sleep(2)


def test_heavy_init_api_no_setup():
    with pytest.warns(RuntimeWarning, match="loading a model or doing other heavy processing inside the constructor"):
        HeavyInitAPI()


def test_heavy_init_api_with_setup(recwarn):
    HeavyInitAPIWithSetup()

    assert len(recwarn) == 0


def test_heavy_setup_api(recwarn):
    api = HeavySetupAPI()
    api.setup("")

    assert len(recwarn) == 0



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Code extraction adapted from https://github.com/tassaron/get_code_from_markdown
import re
import selectors
import subprocess
import sys
import time

import pytest
from tqdm import tqdm

uvicorn_msg = "Uvicorn running on http://0.0.0.0:8000 (Press CTRL+C to quit)"


def extract_code_blocks(lines: list[str]) -> list[str]:
    language = "python"
    regex = re.compile(
        r"(?P<start>^```(?P<block_language>(\w|-)+)\n)(?P<code>.*?\n)(?P<end>```)",
        re.DOTALL | re.MULTILINE,
    )
    blocks = [(match.group("block_language"), match.group("code")) for match in regex.finditer("".join(lines))]
    return [block for block_language, block in blocks if block_language == language]


def get_code_blocks(file: str) -> list[str]:
    with open(file) as f:
        lines = list(f)
        return extract_code_blocks(lines)


def get_extra_time(content: str) -> int:
    if "torch" in content or "transformers" in content:
        return 5

    return 0


def run_script_with_timeout(file, timeout, extra_time, killall):
    sel = selectors.DefaultSelector()
    try:
        process = subprocess.Popen(
            [sys.executable, str(file)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,  # Line-buffered
            universal_newlines=True,  # Decode bytes to string
        )

        stdout_lines = []
        stderr_lines = []
        end_time = time.time() + timeout + extra_time

        sel.register(process.stdout, selectors.EVENT_READ)
        sel.register(process.stderr, selectors.EVENT_READ)

        while True:
            timeout_remaining = end_time - time.time()
            if timeout_remaining <= 0:
                killall(process)
                break

            events = sel.select(timeout=timeout_remaining)
            for key, _ in events:
                if key.fileobj is process.stdout:
                    line = process.stdout.readline()
                    if line:
                        stdout_lines.append(line)
                elif key.fileobj is process.stderr:
                    line = process.stderr.readline()
                    if line:
                        stderr_lines.append(line)

            if process.poll() is not None:
                break

        output = "".join(stdout_lines)
        errors = "".join(stderr_lines)

        # Get the return code of the process
        returncode = process.returncode

    except Exception as e:
        output = ""
        errors = str(e)
        returncode = -1  # Indicate failure in running the process

    return returncode, output, errors


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Windows CI is slow and this test is just a sanity check.")
def test_readme(tmp_path, killall):
    d = tmp_path / "readme_codes"
    d.mkdir(exist_ok=True)
    code_blocks = get_code_blocks("README.md")
    assert len(code_blocks) > 0, "No code block found in README.md"

    for i, code in enumerate(tqdm(code_blocks)):
        file = d / f"{i}.py"
        file.write_text(code)
        extra_time = get_extra_time(code)

        returncode, stdout, stderr = run_script_with_timeout(file, timeout=5, extra_time=extra_time, killall=killall)

        if "server.run" in code:
            assert uvicorn_msg in stderr, f"Expected to run uvicorn server.\nCode:\n {code}\n\nCode output: {stderr}"
        elif "requests.post" in code:
            assert "ConnectionError" in stderr, (
                f"Client examples should fail with a ConnectionError because there is no server running.\nCode:\n{code}"
            )
        else:
            assert returncode == 0, (
                f"Code exited with {returncode}.\n"
                f"Error: {stderr}\n"
                f"Please check the code for correctness:\n```\n{code}\n```"
            )



import subprocess
from pathlib import Path

import pytest


def test_version():
    """Prints the help message for the requirements commands."""
    return_code = subprocess.call(["python", "-mlightning_utilities.cli", "version"])  # noqa: S607
    assert return_code == 0


@pytest.mark.parametrize("args", ["positional", "optional"])
class TestRequirements:
    """Test requirements commands."""

    BASE_CMD = ("python", "-m", "lightning_utilities.cli", "requirements")
    REQUIREMENTS_SAMPLE = """
# This is sample requirements file
#  with multi line comments

torchvision >=0.13.0, <0.16.0  # sample # comment
gym[classic,control] >=0.17.0, <0.27.0
ipython[all] <8.15.0  # strict
torchmetrics >=0.10.0, <1.3.0
deepspeed >=0.8.2, <=0.9.3; platform_system != "Windows"  # strict
    """

    def _create_requirements_file(self, local_path: Path, filename: str = "testing-cli-requirements.txt"):
        """Create a sample requirements file."""
        req_file = local_path / filename
        with open(req_file, "w", encoding="utf8") as fopen:
            fopen.write(self.REQUIREMENTS_SAMPLE)
        return str(req_file)

    def _build_command(self, subcommand: str, cli_params: tuple, arg_style: str):
        """Build the command for the CLI."""
        if arg_style == "positional":
            return list(self.BASE_CMD) + [subcommand] + [value for _, value in cli_params]
        if arg_style == "optional":
            return list(self.BASE_CMD) + [subcommand] + [f"--{key}={value}" for key, value in cli_params]
        raise ValueError(f"Unknown test configuration: {arg_style}")

    def test_requirements_prune_pkgs(self, args, tmp_path):
        """Prune packages from requirements files."""
        req_file = self._create_requirements_file(tmp_path)
        cli_params = (("packages", "ipython"), ("req_files", req_file))
        cmd = self._build_command("prune-pkgs", cli_params, args)
        return_code = subprocess.call(cmd)  # noqa: S603
        assert return_code == 0

    def test_requirements_set_oldest(self, args, tmp_path):
        """Set the oldest version of packages in requirement files."""
        req_file = self._create_requirements_file(tmp_path, "requirements.txt")
        cli_params = (("req_files", req_file),)
        cmd = self._build_command("set-oldest", cli_params, args)
        return_code = subprocess.call(cmd)  # noqa: S603
        assert return_code == 0

    def test_requirements_replace_pkg(self, args, tmp_path):
        """Replace a package in requirements files."""
        req_file = self._create_requirements_file(tmp_path, "requirements.txt")
        cli_params = (("old_package", "torchvision"), ("new_package", "torchtext"), ("req_files", req_file))
        cmd = self._build_command("replace-pkg", cli_params, args)
        return_code = subprocess.call(cmd)  # noqa: S603
        assert return_code == 0



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import io
import os

import numpy as np
from fastapi.testclient import TestClient
from PIL import Image

import litserve as ls
from litserve.schema.image import ImageInput, ImageOutput
from litserve.utils import wrap_litserve_start


class ImageAPI(ls.LitAPI):
    def setup(self, device):
        self.model = lambda x: np.array(x) * 2

    def decode_request(self, request: ImageInput):
        return request.get_image()

    def predict(self, x):
        return self.model(x)

    def encode_response(self, numpy_image) -> ImageOutput:
        output = Image.fromarray(np.uint8(numpy_image)).convert("RGB")
        return ImageOutput(image=output)


def test_image_input_output(tmpdir):
    path = os.path.join(tmpdir, "test.png")
    server = ls.LitServer(ImageAPI(), accelerator="cpu", devices=1, workers_per_device=1)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        Image.new("RGB", (32, 32)).save(path)
        with open(path, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read()).decode("utf-8")
        response = client.post("/predict", json={"image_data": encoded_string})

        assert response.status_code == 200, f"Unexpected status code: {response.status_code}"
        image_data = response.json()["image"]
        image = Image.open(io.BytesIO(base64.b64decode(image_data)))
        assert image.size == (32, 32), f"Unexpected image size: {image.size}"


class MultiImageInputModel(ImageInput):
    image_0: str
    image_1: str
    image_2: str


class MultiImageInputAPI(ImageAPI):
    def decode_request(self, request: MultiImageInputModel):
        images = [request.get_image(f"image_{i}") for i in range(3)]
        for image in images:
            assert isinstance(image, Image.Image)
        return images[0]


def test_multiple_image_input(tmpdir):
    path = os.path.join(tmpdir, "test.png")
    server = ls.LitServer(MultiImageInputAPI(), accelerator="cpu", devices=1, workers_per_device=1)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        data = {}
        for i in range(3):
            Image.new("RGB", (32, 32)).save(path)
            with open(path, "rb") as image_file:
                encoded_string = base64.b64encode(image_file.read()).decode("utf-8")
            data[f"image_{i}"] = encoded_string
        response = client.post("/predict", json=data)

        assert response.status_code == 200, f"Unexpected status code: {response.status_code}"
        image_data = response.json()["image"]
        image = Image.open(io.BytesIO(base64.b64decode(image_data)))
        assert image.size == (32, 32), f"Unexpected image size: {image.size}"



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import contextlib
import json
import sys
import time
from time import sleep
from unittest.mock import MagicMock, patch

import pytest
import torch
import torch.nn as nn
from asgi_lifespan import LifespanManager
from fastapi import HTTPException, Request, Response
from fastapi.testclient import TestClient
from httpx import ASGITransport, AsyncClient

import litserve as ls
from litserve import LitAPI
from litserve.connector import _Connector
from litserve.server import LitServer
from litserve.utils import wrap_litserve_start


def test_index(sync_testclient):
    assert sync_testclient.get("/").text == "litserve running"


@patch("litserve.server.LitServer.lifespan")
def test_device_identifiers(lifespan_mock, simple_litapi):
    server = LitServer(simple_litapi, accelerator="cpu", devices=1, timeout=10)
    assert server.device_identifiers("cpu", 1) == ["cpu:1"]
    assert server.device_identifiers("cpu", [1, 2]) == ["cpu:1", "cpu:2"]

    server = LitServer(simple_litapi, accelerator="cpu", devices=1, timeout=10)
    assert server.devices == ["cpu"]

    server = LitServer(simple_litapi, accelerator="cuda", devices=1, timeout=10)
    assert server.devices == [["cuda:0"]]

    server = LitServer(simple_litapi, accelerator="cuda", devices=[1, 2], timeout=10)
    # [["cuda:1"], ["cuda:2"]]
    assert server.devices[0][0] == "cuda:1"
    assert server.devices[1][0] == "cuda:2"


@pytest.mark.parametrize("devices", ["cpu", ["cpu", "cuda:0"], ["cuda:a", "cuda:1"]])
def test_device_identifiers_error(simple_litapi, devices):
    with pytest.raises(
        ValueError, match="devices must be an integer or a list of integers when using 'cuda' or 'mps', instead got .*"
    ):
        LitServer(simple_litapi, accelerator="cuda", devices=devices, timeout=10)


@pytest.mark.parametrize("use_zmq", [True, False])
@pytest.mark.asyncio
async def test_stream(simple_stream_api, use_zmq):
    simple_stream_api.stream = True
    server = LitServer(simple_stream_api, timeout=10, fast_queue=use_zmq)
    expected_output1 = "prompt=Hello generated_output=LitServe is streaming output".lower().replace(" ", "")
    expected_output2 = "prompt=World generated_output=LitServe is streaming output".lower().replace(" ", "")

    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            # TODO: remove this sleep when we have a better way to check if the server is ready
            # TODO: main process can only consume when response_queue_to_buffer is ready
            await asyncio.sleep(4)
            resp1 = ac.post("/predict", json={"prompt": "Hello"}, timeout=10)
            resp2 = ac.post("/predict", json={"prompt": "World"}, timeout=10)
            resp1, resp2 = await asyncio.gather(resp1, resp2)
            assert resp1.status_code == 200, "Check if server is running and the request format is valid."
            assert resp1.text == expected_output1, (
                "Server returns input prompt and generated output which didn't match."
            )
            assert resp2.status_code == 200, "Check if server is running and the request format is valid."
            assert resp2.text == expected_output2, (
                "Server returns input prompt and generated output which didn't match."
            )


@pytest.mark.parametrize("use_zmq", [True, False])
@pytest.mark.asyncio
async def test_batched_stream_server(simple_batched_stream_api, use_zmq):
    api = simple_batched_stream_api
    api.stream = True
    api.max_batch_size = 4
    api.batch_timeout = 2
    server = LitServer(api, timeout=30, fast_queue=use_zmq)
    expected_output1 = "Hello LitServe is streaming output".lower().replace(" ", "")
    expected_output2 = "World LitServe is streaming output".lower().replace(" ", "")

    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp1 = ac.post("/predict", json={"prompt": "Hello"}, timeout=10)
            resp2 = ac.post("/predict", json={"prompt": "World"}, timeout=10)
            resp1, resp2 = await asyncio.gather(resp1, resp2)
            assert resp1.status_code == 200, "Check if server is running and the request format is valid."
            assert resp2.status_code == 200, "Check if server is running and the request format is valid."
            assert resp1.text == expected_output1, (
                "Server returns input prompt and generated output which didn't match."
            )
            assert resp2.text == expected_output2, (
                "Server returns input prompt and generated output which didn't match."
            )


def test_litapi_with_stream(simple_litapi_cls):
    with pytest.raises(
        ValueError,
        match="""When `stream=True` both `lit_api.predict` and
             `lit_api.encode_response` must generate values using `yield""",
    ):
        # TODO: The error should be raised in the LitAPI constructor
        LitServer(simple_litapi_cls(stream=True))


class Linear(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(1, 1)
        self.linear.weight.data.fill_(2.0)
        self.linear.bias.data.fill_(1.0)

    def forward(self, x):
        return self.linear(x)


class SimpleLitAPI(LitAPI):
    def setup(self, device):
        self.model = Linear().to(device)
        self.device = device

    def decode_request(self, request: Request):
        content = request["input"]
        return torch.tensor([content], device=self.device)

    def predict(self, x):
        return self.model(x[None, :])

    def encode_response(self, output) -> Response:
        return {"output": float(output)}


@pytest.mark.parametrize(
    ("input_accelerator", "expected_accelerator"),
    [
        ("cpu", "cpu"),
        pytest.param(
            "cuda",
            "cuda",
            marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU"),
        ),
        pytest.param(
            None, "cuda", marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU")
        ),
        pytest.param(
            "auto",
            "cuda",
            marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU"),
        ),
        pytest.param(
            "auto",
            "mps",
            marks=pytest.mark.skipif(not torch.backends.mps.is_available(), reason="Only tested on Apple MPS"),
        ),
        pytest.param(
            None,
            "mps",
            marks=pytest.mark.skipif(not torch.backends.mps.is_available(), reason="Only tested on Apple MPS"),
        ),
    ],
)
def test_auto_accelerator(input_accelerator, expected_accelerator):
    server = LitServer(SimpleLitAPI(), devices=1, timeout=10, accelerator=input_accelerator)
    assert server._connector.accelerator == expected_accelerator


def test_mocked_accelerator():
    # 1. cuda available
    with patch("litserve.connector.check_cuda_with_nvidia_smi", return_value=True):
        connector = _Connector(accelerator="auto")
        assert connector.accelerator == "cuda"

    # 2. mps available
    with patch("litserve.connector._Connector._choose_gpu_accelerator_backend", return_value="mps"):
        server = LitServer(SimpleLitAPI(), devices=1, timeout=10, accelerator="auto")
        assert server._connector.accelerator == "mps"


@patch("litserve.server.uvicorn")
def test_server_run(mock_uvicorn, mock_manager):
    server = LitServer(SimpleLitAPI())
    server.verify_worker_status = MagicMock()
    with pytest.raises(ValueError, match="port must be a value from 1024 to 65535 but got"):
        server.run(port="invalid port")

    with pytest.raises(ValueError, match="port must be a value from 1024 to 65535 but got"):
        server.run(port=65536)

    with pytest.raises(ValueError, match="host must be '0.0.0.0', '127.0.0.1', or '::' but got"):
        server.run(host="127.0.0.2")

    # test port 8000
    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(port=8000)
    mock_uvicorn.Config.assert_called()
    mock_uvicorn.reset_mock()

    # test port 8001
    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(port="8001")
    mock_uvicorn.Config.assert_called()

    # test host "::" and port 8000
    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(host="::", port="8000")
    mock_uvicorn.Config.assert_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Test is only for Unix")
@patch("litserve.server.uvicorn")
def test_start_server(mock_uvicon):
    api = ls.test_examples.TestAPI(spec=ls.OpenAISpec())
    server = LitServer(api)
    sockets = MagicMock()
    server._start_server(8000, 1, "info", sockets, "process")
    mock_uvicon.Server.assert_called()
    assert server.lit_api.spec.response_queue_id is not None, "response_queue_id must be generated"


@pytest.fixture
def server_for_api_worker_test(simple_litapi):
    server = ls.LitServer(simple_litapi, devices=1)
    server.verify_worker_status = MagicMock()
    server.launch_inference_worker = MagicMock(return_value=[MagicMock()])
    server._start_server = MagicMock()
    server._transport = MagicMock()
    return server


@pytest.mark.skipif(sys.platform == "win32", reason="Test is only for Unix")
@patch("litserve.server.uvicorn")
def test_server_run_with_api_server_worker_type(mock_uvicorn, server_for_api_worker_test, mock_manager):
    server = server_for_api_worker_test

    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(api_server_worker_type="process", num_api_servers=10)
    server.launch_inference_worker.assert_called_with(server.lit_api)


@pytest.mark.skipif(sys.platform == "win32", reason="Test is only for Unix")
@pytest.mark.parametrize(("api_server_worker_type", "num_api_workers"), [(None, 1), ("process", 1)])
@patch("litserve.server.uvicorn")
def test_server_run_with_process_api_worker(
    mock_uvicorn, api_server_worker_type, num_api_workers, server_for_api_worker_test, mock_manager
):
    server = server_for_api_worker_test

    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(api_server_worker_type=api_server_worker_type, num_api_workers=num_api_workers)
    server.launch_inference_worker.assert_called_with(server.lit_api)
    actual = server._start_server.call_args
    assert actual[0][4] == "process", "Server should run in process mode"
    mock_uvicorn.Config.assert_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Test is only for Unix")
@patch("litserve.server.uvicorn")
def test_server_run_with_thread_api_worker(mock_uvicorn, server_for_api_worker_test, mock_manager):
    server = server_for_api_worker_test
    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(api_server_worker_type="thread")
    server.launch_inference_worker.assert_called_with(server.lit_api)
    assert server._start_server.call_args[0][4] == "thread", "Server should run in thread mode"
    mock_uvicorn.Config.assert_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Test is only for Unix")
def test_server_run_with_invalid_api_worker(simple_litapi):
    server = ls.LitServer(simple_litapi, devices=1)
    server.verify_worker_status = MagicMock()
    with pytest.raises(ValueError, match=r"Must be 'process' or 'thread'"):
        server.run(api_server_worker_type="invalid")

    with pytest.raises(ValueError, match=r"must be greater than 0"):
        server.run(num_api_servers=0)


@pytest.mark.skipif(sys.platform != "win32", reason="Test is only for Windows")
@patch("litserve.server.uvicorn")
def test_server_run_windows(mock_uvicorn, mock_manager):
    api = ls.test_examples.SimpleLitAPI()
    server = ls.LitServer(api)
    server.verify_worker_status = MagicMock()
    server.launch_inference_worker = MagicMock(return_value=[MagicMock()])
    server._transport = MagicMock()
    server._start_server = MagicMock()

    with patch("litserve.server.mp.Manager", return_value=mock_manager):
        server.run(api_server_worker_type=None)
    actual = server._start_server.call_args
    assert actual[0][4] == "thread", "Windows only supports thread mode"


def test_server_terminate():
    server = LitServer(SimpleLitAPI())
    server.verify_worker_status = MagicMock()
    server._transport = MagicMock()

    with (
        patch("litserve.server.LitServer._init_manager", return_value=MagicMock()) as mock_init_manager,
        patch("litserve.server.LitServer._start_server", side_effect=Exception("mocked error")) as mock_start,
        patch("litserve.server.LitServer.launch_inference_worker", return_value=([MagicMock()])) as mock_launch,
    ):
        with pytest.raises(Exception, match="mocked error"):
            server.run(port=8001)

        mock_init_manager.assert_called()
        mock_launch.assert_called()
        mock_start.assert_called()
        server._transport.close.assert_called()


@pytest.mark.parametrize(("disable_openapi_url", "should_print"), [(False, True), (True, False)])
@patch("builtins.print")
@patch("litserve.server.uvicorn")
def test_disable_openapi_url_print_message(mock_uvicorn, mock_print, mock_manager, disable_openapi_url, should_print):
    """Test that the Swagger UI message is only printed when disable_openapi_url=False."""
    server = LitServer(SimpleLitAPI(), disable_openapi_url=disable_openapi_url)
    server.verify_worker_status = MagicMock()

    with (
        patch("litserve.server.mp.Manager", return_value=mock_manager),
        contextlib.suppress(Exception),
    ):
        server.run(port=8000)

    if should_print:
        mock_print.assert_called_with("Swagger UI is available at http://0.0.0.0:8000/docs")
    else:
        mock_print.assert_not_called()


class IdentityAPI(ls.test_examples.SimpleLitAPI):
    def predict(self, x, context):
        context["input"] = x
        return self.model(x)

    def encode_response(self, output, context):
        input = context["input"]
        return {"output": input}


class IdentityBatchedAPI(ls.test_examples.SimpleBatchedAPI):
    def predict(self, x_batch, context):
        for c, x in zip(context, x_batch):
            c["input"] = x
        return self.model(x_batch)

    def encode_response(self, output, context):
        input = context["input"]
        return {"output": input}


class IdentityBatchedStreamingAPI(ls.test_examples.SimpleBatchedAPI):
    def predict(self, x_batch, context):
        for c, x in zip(context, x_batch):
            c["input"] = x
        yield self.model(x_batch)

    def encode_response(self, output_stream, context):
        for _ in output_stream:
            yield [{"output": ctx["input"]} for ctx in context]


class PredictErrorAPI(ls.test_examples.SimpleLitAPI):
    def predict(self, x, y, context):
        context["input"] = x
        return self.model(x)

    def encode_response(self, output, context):
        input = context["input"]
        return {"output": input}


@pytest.mark.asyncio
async def test_inject_context():
    # Test context injection with single loop
    api = IdentityAPI()
    server = LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp = await ac.post("/predict", json={"input": 5.0}, timeout=10)
    assert resp.json()["output"] == 5.0, "output from Identity server must be same as input"

    # Test context injection with batched loop
    server = LitServer(IdentityBatchedAPI(max_batch_size=2, batch_timeout=0.01))
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp = await ac.post("/predict", json={"input": 5.0}, timeout=10)
    assert resp.json()["output"] == 5.0, "output from Identity server must be same as input"

    # Test context injection with batched streaming loop
    server = LitServer(IdentityBatchedStreamingAPI(max_batch_size=2, batch_timeout=0.01, stream=True))
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp = await ac.post("/predict", json={"input": 5.0}, timeout=10)
    assert resp.json()["output"] == 5.0, "output from Identity server must be same as input"

    server = LitServer(PredictErrorAPI())
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        resp = client.post("/predict", json={"input": 5.0}, timeout=10)
        assert resp.status_code == 500, "predict() missed 1 required positional argument: 'y'"


def test_custom_api_path():
    with pytest.raises(ValueError, match="api_path must start with '/'"):
        LitServer(ls.test_examples.SimpleLitAPI(api_path="predict"))

    server = LitServer(ls.test_examples.SimpleLitAPI(api_path="/v1/predict"))
    assert "/v1/predict" in [route.path for route in server.app.routes]


def test_custom_healthcheck_path():
    with pytest.raises(ValueError, match="healthcheck_path must start with '/'. "):
        LitServer(ls.test_examples.SimpleLitAPI(), healthcheck_path="customhealth")

    server = LitServer(ls.test_examples.SimpleLitAPI(), healthcheck_path="/v1/custom_health")
    url = server.healthcheck_path
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        # Sleep a bit to ensure the server is ready
        sleep(3)
        response = client.get(url)

    assert response.status_code == 200, "Server response should be 200 (OK)"


def test_custom_info_path():
    with pytest.raises(ValueError, match="info_path must start with '/'. "):
        LitServer(ls.test_examples.SimpleLitAPI(), info_path="custominfo")

    server = LitServer(ls.test_examples.SimpleLitAPI(), info_path="/v1/custom_info", accelerator="cpu")
    url = server.info_path
    expected_response = {
        "model": None,
        "server": {
            "devices": ["cpu"],
            "workers_per_device": 1,
            "timeout": 30,
            "stream": {"/predict": False},
            "max_payload_size": None,
            "track_requests": False,
        },
    }

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        # Sleep a bit to ensure the server is ready
        sleep(3)
        response = client.get(url)

    assert response.status_code == 200, "Server response should be 200 (OK)"
    assert response.json() == expected_response, "server didn't return expected output"


def test_info_route():
    model_metadata = {"name": "my-awesome-model", "version": "v1.1.0"}
    expected_response = {
        "model": {
            "name": "my-awesome-model",
            "version": "v1.1.0",
        },
        "server": {
            "devices": ["cpu"],
            "workers_per_device": 1,
            "timeout": 30,
            "stream": {"/predict": False},
            "max_payload_size": None,
            "track_requests": False,
        },
    }

    server = ls.LitServer(ls.test_examples.SimpleLitAPI(), accelerator="cpu", model_metadata=model_metadata)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.get("/info", headers={"Host": "localhost"})
        assert response.status_code == 200, f"Expected response to be 200 but got {response.status_code}"
        assert response.json() == expected_response, "server didn't return expected output"


def test_model_metadata_json_error():
    with pytest.raises(ValueError, match="model_metadata must be JSON serializable"):
        ls.LitServer(ls.test_examples.SimpleLitAPI(), model_metadata=int)


class TestHTTPExceptionAPI(ls.test_examples.SimpleLitAPI):
    def decode_request(self, request):
        raise HTTPException(501, "decode request is bad")


class TestHTTPExceptionAPI2(ls.test_examples.SimpleLitAPI):
    def decode_request(self, request):
        raise HTTPException(status_code=400, detail="decode request is bad")


def test_http_exception():
    server = LitServer(TestHTTPExceptionAPI(), fast_queue=True)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.status_code == 501, "Server raises 501 error"
        assert response.text == '{"detail":"decode request is bad"}', "decode request is bad"

    server = LitServer(TestHTTPExceptionAPI2(), fast_queue=True)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.status_code == 400, "Server raises 400 error"
        assert response.text == '{"detail":"decode request is bad"}', "decode request is bad"


def test_generate_client_file(tmp_path, monkeypatch):
    expected = """import requests

response = requests.post("http://127.0.0.1:8123/predict", json={"input": 4.0})
print(f"Status: {response.status_code}\\nResponse:\\n {response.text}")"""
    monkeypatch.chdir(tmp_path)
    LitServer.generate_client_file(8123)
    with open(tmp_path / "client.py") as fr:
        assert expected in fr.read(), f"Expected {expected} in client.py"

    LitServer.generate_client_file(8000)
    with open(tmp_path / "client.py") as fr:
        assert expected in fr.read(), "Shouldn't replace existing client.py"


class FailFastAPI(ls.test_examples.SimpleLitAPI):
    def setup(self, device):
        raise ValueError("setup failed")


@pytest.mark.parametrize("use_zmq", [True, False])
def test_workers_setup_status(use_zmq, port):
    server = LitServer(
        FailFastAPI(),
        fast_queue=use_zmq,
    )
    with pytest.raises(RuntimeError, match="One or more workers failed to start. Shutting down LitServe"):
        server.run(port=port, log_level="error")


def test_max_batch_size_warning(simple_litapi):
    # Remove this test after v0.3.0
    with pytest.warns(DeprecationWarning, match="'max_batch_size' and 'batch_timeout' are being deprecated"):
        ls.LitServer(simple_litapi, max_batch_size=4)


class TestAsyncLitAPI(ls.LitAPI):
    def setup(self, device):
        self.model = None

    async def decode_request(self, request):
        return request["input"]

    async def predict(self, x):
        return x**2

    async def encode_response(self, output):
        return {"output": output}


@pytest.mark.asyncio
async def test_async_litapi():
    api = TestAsyncLitAPI(enable_async=True)
    server = LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp = await ac.post("/predict", json={"input": 5.0}, timeout=10)
            assert resp.status_code == 200, "Server response should be 200 (OK)"
            assert resp.json()["output"] == 25.0, "output from Identity server must be same as input"


class TestSleepAsyncLitAPI(TestAsyncLitAPI):
    async def predict(self, x):
        # simulate a long-running task
        await asyncio.sleep(4)
        return x**2


@pytest.mark.asyncio
@pytest.mark.parametrize("num_requests", [10, 50, 100])
async def test_concurrent_async_inference(num_requests):
    """Test that async API processes requests concurrently (not sequentially)."""
    api = TestSleepAsyncLitAPI(enable_async=True)
    server = LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            sleep(2)  # Sleep a bit to ensure the server is ready

            tasks = [ac.post("/predict", json={"input": 5.0}, timeout=10) for _ in range(num_requests)]
            start = time.perf_counter()
            responses = await asyncio.gather(*tasks)
            elapsed = time.perf_counter() - start

            for resp in responses:
                assert resp.status_code == 200
                assert resp.json()["output"] == 25.0

            # All requests should finish in just over 4s, plus some overhead
            assert elapsed < 4 + 4, f"Expected all requests to finish in just over 4s, but took {elapsed:.2f}s."


class TestHTTPExceptionAsyncLitAPI(TestAsyncLitAPI):
    async def decode_request(self, request):
        raise HTTPException(status_code=501, detail="decode request is bad")


@pytest.mark.asyncio
async def test_error_propagation_in_async_litapi():
    server = LitServer(TestHTTPExceptionAsyncLitAPI(enable_async=True))
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp = await ac.post("/predict", json={"input": 5.0}, timeout=10)
            assert resp.status_code == 501, "Server raises 501 error"
            assert resp.json() == {"detail": "decode request is bad"}, "decode request is bad"


class TestAsyncStreamLitAPI(ls.LitAPI):
    def setup(self, device):
        self.model = lambda x: x

    async def decode_request(self, request):
        return request["input"]

    async def predict(self, x):
        for i in range(5):
            yield self.model(i)

    async def encode_response(self, output_stream):
        async for output in output_stream:
            yield {"output": output}


@pytest.mark.asyncio
async def test_async_stream_litapi():
    api = TestAsyncStreamLitAPI(enable_async=True, stream=True)
    server = LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            resp = await ac.post("/predict", json={"input": 4.0}, timeout=10)
            assert resp.status_code == 200, "Server response should be 200 (OK)"
            chunks = []
            async for line in resp.aiter_lines():
                if line.strip():
                    chunks.append(json.loads(line)["output"])
            assert len(chunks) == 5, "Expected 5 chunks of output"
            assert chunks == list(range(5)), "Expected output to be a sequence of integers from 0 to 4"


class TestSleepAsyncStreamLitAPI(TestAsyncStreamLitAPI):
    async def predict(self, x):
        for i in range(5):
            await asyncio.sleep(0.5)  # Simulate streaming delay
            yield i


@pytest.mark.asyncio
@pytest.mark.parametrize("num_requests", [10, 50, 100])
async def test_concurrent_async_streaming_inference(num_requests):
    api = TestSleepAsyncStreamLitAPI(enable_async=True, stream=True)
    server = LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            sleep(2)  # Sleep a bit to ensure the server is ready

            # Prepare concurrent streaming requests using the parameterized num_requests
            tasks = [ac.post("/predict", json={"input": 4.0}, timeout=10) for _ in range(num_requests)]
            start = time.perf_counter()
            responses = await asyncio.gather(*tasks)
            elapsed = time.perf_counter() - start

            # Collect and check streamed outputs for each response
            for resp in responses:
                assert resp.status_code == 200, "Server response should be 200 (OK)"
                chunks = []
                async for line in resp.aiter_lines():
                    if line.strip():
                        chunks.append(json.loads(line)["output"])
                assert len(chunks) == 5, "Expected 5 chunks of output"
                assert chunks == list(range(5)), "Expected output to be a sequence of integers from 0 to 4"

            # All requests should finish in just over the streaming time for one request
            assert elapsed < 4 + 4, (
                f"Expected all requests to finish in just over 4s, plus some overhead, but took {elapsed:.2f}s."
            )



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import inspect
import json
import warnings
from abc import ABC
from collections.abc import Awaitable
from queue import Queue
from typing import TYPE_CHECKING, Callable, Optional, Union

from pydantic import BaseModel

from litserve.specs.base import LitSpec
from litserve.utils import _TimedInitMeta

if TYPE_CHECKING:
    from litserve.loops.base import LitLoop
    from litserve.mcp import MCP


class LitAPI(ABC, metaclass=_TimedInitMeta):
    """Define inference logic for the model.

    LitAPI is the core abstraction for serving AI models with LitServe. It provides a clean
    interface for model loading, request processing, and response generation with automatic
    optimizations like batching, streaming, and async processing.

    Core Workflow:
        1. **setup()**: Load and initialize the model once per worker
        2. **decode_request()**: Convert HTTP request to model input format
        3. **predict()**: Run model inference on the input
        4. **encode_response()**: Convert model output to HTTP response format

    Quick Start:
        ```python
        import litserve as ls

        class MyAPI(ls.LitAPI):
            def setup(self, device):
                self.model = lambda x: x**2

            def predict(self, x):
                return self.model(x["input"])

        server = ls.LitServer(MyAPI())
        server.run()
        ```

    Required Methods:
        setup(device): Initialize the model and resources
        predict(x): Core inference logic

    Optional Methods:
        decode_request(request): Transform HTTP requests to model input
        encode_response(output): Transform model outputs to HTTP responses
        batch(inputs)/unbatch(outputs): Custom batching logic

    Configuration:
        max_batch_size: Batch multiple requests for better GPU utilization. Defaults to 1.
        batch_timeout: Wait time for batch to fill (seconds). Defaults to 0.0.
        stream: Enable streaming responses for real-time output. Defaults to False.
        api_path: URL endpoint path. Defaults to "/predict".
        enable_async: Enable async/await for non-blocking operations. Defaults to False.
        spec: API specification (e.g., OpenAISpec for OpenAI compatibility). Defaults to None.
        mcp: Model Context Protocol integration for AI assistants. Defaults to None.

    Examples:
        Batched GPU Inference:
        ```python
        class BatchedAPI(ls.LitAPI):
            def setup(self, device):
                self.model = load_model().to(device)

            def predict(self, batch):
                return self.model(batch)

        api = BatchedAPI(max_batch_size=8, batch_timeout=0.1)
        ```

        Streaming LLM:
        ```python
        class StreamingLLM(ls.LitAPI):
            def setup(self, device):
                self.model = load_llm()

            def predict(self, prompt):
                for token in self.model.generate_stream(prompt):
                    yield token

        api = StreamingLLM(stream=True)
        ```

        OpenAI-Compatible:
        ```python
        from litserve.specs import OpenAISpec

        class ChatAPI(ls.LitAPI):
            def setup(self, device):
                self.model = load_chat_model()

            def predict(self, messages):
                return self.model.chat(messages)

        api = ChatAPI(spec=OpenAISpec())
        ```

    Performance Tips:
        - Use batching for GPU models to maximize utilization
        - Enable streaming for operations taking >1 second
        - Use async for I/O-bound operations (databases, external APIs)
        - Load models in setup(), not __init__
        - Monitor GPU memory usage with larger batch sizes

    See Also:
        - LitServer: Server class for hosting APIs
        - LitSpec: API specifications for standard interfaces

    """

    _stream: bool = False
    _default_unbatch: Optional[Callable] = None
    _spec: Optional[LitSpec] = None
    _device: Optional[str] = None
    _logger_queue: Optional[Queue] = None
    request_timeout: Optional[float] = None

    def __init__(
        self,
        max_batch_size: int = 1,
        batch_timeout: float = 0.0,
        api_path: str = "/predict",
        stream: bool = False,
        loop: Optional[Union[str, "LitLoop"]] = "auto",
        spec: Optional[LitSpec] = None,
        mcp: Optional["MCP"] = None,
        enable_async: bool = False,
    ):
        """Initialize LitAPI with configuration options."""

        if max_batch_size <= 0:
            raise ValueError("max_batch_size must be greater than 0")

        if batch_timeout < 0:
            raise ValueError("batch_timeout must be greater than or equal to 0")

        if isinstance(spec, LitSpec):
            stream = spec.stream

        if loop is None:
            loop = "auto"

        if isinstance(loop, str) and loop != "auto":
            raise ValueError("loop must be an instance of _BaseLoop or 'auto'")

        if not api_path.startswith("/"):
            raise ValueError(
                "api_path must start with '/'. "
                "Please provide a valid api path like '/predict', '/classify', or '/v1/predict'"
            )

        # Check if the batch and unbatch methods are overridden in the lit_api instance
        batch_overridden = self.batch.__code__ is not LitAPI.batch.__code__
        unbatch_overridden = self.unbatch.__code__ is not LitAPI.unbatch.__code__

        if batch_overridden and unbatch_overridden and max_batch_size == 1:
            warnings.warn(
                "The LitServer has both batch and unbatch methods implemented, "
                "but the max_batch_size parameter was not set."
            )

        self._api_path = api_path
        self.stream = stream
        self._loop = loop
        self._spec = spec
        self.max_batch_size = max_batch_size
        self.batch_timeout = batch_timeout
        self.enable_async = enable_async
        self._validate_async_methods()
        self.mcp = mcp
        if mcp:
            mcp._connect(self)

    def _validate_async_methods(self):
        """Validate that async methods are properly implemented when enable_async is True."""
        if not self.enable_async:
            return

        # Define validation rules for each method
        validation_rules = {
            "decode_request": {
                "required_types": [asyncio.iscoroutinefunction, inspect.isasyncgenfunction],
                "error_type": "warning",
                "message": "should be an async function or async generator when enable_async=True",
            },
            "encode_response": {
                "required_types": [asyncio.iscoroutinefunction, inspect.isasyncgenfunction],
                "error_type": "warning",
                "message": "should be an async function or async generator when enable_async=True",
            },
            "predict": {
                "required_types": [inspect.isasyncgenfunction, asyncio.iscoroutinefunction],
                "error_type": "error",
                "message": "must be an async generator or async function when enable_async=True",
            },
        }

        errors = []
        warnings_list = []

        for method_name, rules in validation_rules.items():
            method_obj = getattr(self, method_name)

            # Check if method satisfies any of the required types
            is_valid = any(check_func(method_obj) for check_func in rules["required_types"])

            if not is_valid:
                message = f"{method_name} {rules['message']}"

                if rules["error_type"] == "error":
                    errors.append(message)
                else:
                    warnings_list.append(message)

        # Emit warnings
        for warning_msg in warnings_list:
            warnings.warn(f"{warning_msg}. LitServe will asyncify the method.", UserWarning)

        # Raise errors if any
        if errors:
            error_msg = "Async validation failed:\n" + "\n".join(f"- {err}" for err in errors)
            raise ValueError(error_msg)

    def setup(self, device):
        """Setup the model so it can be called in `predict`."""
        pass

    def decode_request(self, request, **kwargs):
        """Convert the request payload to your model input."""
        if self._spec:
            return self._spec.decode_request(request, **kwargs)
        return request

    def batch(self, inputs):
        """Convert a list of inputs to a batched input."""
        # consider assigning an implementation when starting server
        # to avoid the runtime cost of checking (should be negligible)
        if hasattr(inputs[0], "__torch_function__"):
            import torch

            return torch.stack(inputs)
        if inputs[0].__class__.__name__ == "ndarray":
            import numpy

            return numpy.stack(inputs)

        return inputs

    def predict(self, x, **kwargs):
        """Run the model on the input and return or yield the output."""
        raise NotImplementedError("predict is not implemented")

    def _unbatch_no_stream(self, output):
        if isinstance(output, str):
            warnings.warn(
                "The 'predict' method returned a string instead of a list of predictions. "
                "When batching is enabled, 'predict' must return a list to handle multiple inputs correctly. "
                "Please update the 'predict' method to return a list of predictions to avoid unexpected behavior.",
                UserWarning,
            )
        return list(output)

    def _unbatch_stream(self, output_stream):
        for output in output_stream:
            yield list(output)

    def unbatch(self, output):
        """Convert a batched output to a list of outputs."""
        if self._default_unbatch is None:
            raise ValueError(
                "Default implementation for `LitAPI.unbatch` method was not found. "
                "Please implement the `LitAPI.unbatch` method."
            )
        return self._default_unbatch(output)

    def encode_response(self, output, **kwargs):
        """Convert the model output to a response payload.

        To enable streaming, it should yield the output.

        """
        if self._spec:
            return self._spec.encode_response(output, **kwargs)
        return output

    def format_encoded_response(self, data):
        if isinstance(data, dict):
            return json.dumps(data) + "\n"
        if isinstance(data, BaseModel):
            return data.model_dump_json() + "\n"
        return data

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, value):
        self._stream = value

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, value):
        self._device = value

    def pre_setup(self, spec: Optional[LitSpec] = None):
        spec = spec or self._spec
        if self.stream:
            self._default_unbatch = self._unbatch_stream
        else:
            self._default_unbatch = self._unbatch_no_stream

        if spec:
            self._spec = spec
            spec._max_batch_size = self.max_batch_size
            spec.pre_setup(self)

    def set_logger_queue(self, queue: Queue):
        """Set the queue for logging events."""

        self._logger_queue = queue

    def log(self, key, value):
        """Log a key-value pair to the server."""
        if self._logger_queue is None:
            warnings.warn(
                f"Logging event ('{key}', '{value}') attempted without a configured logger. "
                "To track and visualize metrics, please initialize and attach a logger. "
                "If this is intentional, you can safely ignore this message."
            )
            return
        self._logger_queue.put((key, value))

    def has_active_requests(self) -> bool:
        raise NotImplementedError("has_active_requests is not implemented")

    def has_capacity(self) -> bool:
        raise NotImplementedError("has_capacity is not implemented")

    def health(self) -> Union[bool, Awaitable[bool]]:
        """Check the additional health status of the API.

        This method is used in the /health endpoint of the server to determine the health status.
        Users can extend this method to include additional health checks specific to their application.

        The default implementation is synchronous but users may optionally implement an
        ``async`` version which will be awaited by :class:`~litserve.server.LitServer`.

        Returns:
            bool: ``True`` if the API is healthy, ``False`` otherwise.

        """
        return True

    @property
    def loop(self):
        if self._loop == "auto":
            from litserve.loops.loops import get_default_loop

            self._loop = get_default_loop(self.stream, self.max_batch_size, self.enable_async)
        return self._loop

    @loop.setter
    def loop(self, value: "LitLoop"):
        self._loop = value

    @property
    def spec(self):
        return self._spec

    @spec.setter
    def spec(self, value: LitSpec):
        self._spec = value

    @property
    def api_path(self):
        if self._spec:
            return self._spec.api_path
        return self._api_path

    @api_path.setter
    def api_path(self, value: str):
        self._api_path = value



from pathlib import Path

from lightning_utilities.cli.dependencies import (
    prune_packages_in_requirements,
    replace_oldest_version,
    replace_package_in_requirements,
)

_PATH_ROOT = Path(__file__).parent.parent.parent


def test_prune_packages_txt(tmpdir):
    req_file = tmpdir / "requirements.txt"
    with open(req_file, "w") as fp:
        fp.writelines(["fire\n", "abc>=0.1\n"])
    prune_packages_in_requirements("abc", req_files=[str(req_file)])
    with open(req_file) as fp:
        lines = fp.readlines()
    assert lines == ["fire\n"]


def test_oldest_packages_txt(tmpdir):
    req_file = tmpdir / "requirements.txt"
    with open(req_file, "w") as fp:
        fp.writelines(["fire>0.2\n", "abc>=0.1\n"])
    replace_oldest_version(req_files=[str(req_file)])
    with open(req_file) as fp:
        lines = fp.readlines()
    assert lines == ["fire>0.2\n", "abc==0.1\n"]


def test_replace_packages_txt(tmpdir):
    req_file = tmpdir / "requirements.txt"
    with open(req_file, "w") as fp:
        fp.writelines(["torchvision>=0.2\n", "torch>=1.0 # comment\n", "torchtext <0.3\n"])
    replace_package_in_requirements(old_package="torch", new_package="pytorch", req_files=[str(req_file)])
    with open(req_file) as fp:
        lines = fp.readlines()
    assert lines == ["torchvision>=0.2\n", "pytorch>=1.0 # comment\n", "torchtext <0.3\n"]


def test_oldest_packages_pyproject_toml(tmpdir):
    req_file = tmpdir / "pyproject.toml"
    with open(req_file, "w") as fp:
        fp.writelines([
            "[project]\n",
            "dependencies = [\n",
            '    "fire>0.2",\n',
            '    "abc>=0.1",\n',
            "]\n",
        ])
    replace_oldest_version(req_files=str(req_file))
    with open(req_file) as fp:
        lines = fp.readlines()
    assert lines == [
        "[project]\n",
        "dependencies = [\n",
        '    "fire>0.2",\n',
        '    "abc==0.1",\n',
        "]\n",
    ]


def test_replace_packages_pyproject_toml(tmpdir):
    req_file = tmpdir / "pyproject.toml"
    with open(req_file, "w") as fp:
        fp.writelines([
            "[project]\n",
            "dependencies = [\n",
            '    "fire>0.2",\n',
            '    "abc>=0.1",\n',
            "]\n",
        ])
    replace_package_in_requirements(req_files=str(req_file), old_package="fire", new_package="water")
    with open(req_file) as fp:
        lines = fp.readlines()
    assert lines == [
        "[project]\n",
        "dependencies = [\n",
        '    "water>0.2",\n',
        '    "abc>=0.1",\n',
        "]\n",
    ]



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import abstractmethod
from collections.abc import AsyncGenerator, Generator
from typing import TYPE_CHECKING, Callable, Optional, Union

if TYPE_CHECKING:
    from litserve import LitAPI, LitServer


class LitSpec:
    """Spec will have its own encode, and decode."""

    def __init__(self):
        self._endpoints = []
        self.api_path = None
        self._server: LitServer = None
        self._max_batch_size = 1
        self.response_buffer = None
        self.request_queue = None
        self.response_queue_id = None

    @property
    def stream(self):
        return False

    def pre_setup(self, lit_api: "LitAPI"):
        pass

    def setup(self, server: "LitServer"):
        """This method is called by the server to connect the spec to the server."""
        self.response_buffer = server.response_buffer
        self.request_queue = server._get_request_queue(self.api_path)
        self.data_streamer = server.data_streamer

    def add_endpoint(self, path: str, endpoint: Callable, methods: list[str]):
        """Register an endpoint in the spec."""
        self._endpoints.append((path, endpoint, methods))

    @property
    def endpoints(self):
        return self._endpoints.copy()

    @abstractmethod
    def decode_request(self, request, meta_kwargs):
        """Convert the request payload to your model input."""
        pass

    @abstractmethod
    def encode_response(self, output, meta_kwargs):
        """Convert the model output to a response payload.

        To enable streaming, it should yield the output.

        """
        pass

    def as_async(self) -> "_AsyncSpecWrapper":
        return _AsyncSpecWrapper(self)


class _AsyncSpecWrapper:
    def __init__(self, spec: LitSpec):
        self._spec = spec

    def __getattr__(self, name):
        # Delegate all other attributes/methods to the wrapped spec
        return getattr(self._spec, name)

    async def decode_request(self, request, context_kwargs: Optional[dict] = None):
        return self._spec.decode_request(request, context_kwargs)

    async def encode_response(self, output: Union[Generator, AsyncGenerator], context_kwargs: Optional[dict] = None):
        return self._spec.encode_response(output, context_kwargs)



import litserve as ls
from litserve import OpenAIEmbeddingSpec
from litserve.test_examples.openai_embedding_spec_example import TestEmbedAPI

if __name__ == "__main__":
    api = TestEmbedAPI(spec=OpenAIEmbeddingSpec())
    server = ls.LitServer(api, fast_queue=True)
    server.run()



# Licensed under the Apache License, Version 2.0 (the "License");
#     http://www.apache.org/licenses/LICENSE-2.0
#
"""Adjusting version across PTorch ecosystem."""

import logging
import os
import re
import sys
from typing import Optional


def _determine_torchaudio(torch_version: str) -> str:
    """Determine the torchaudio version based on the torch version.

    >>> _determine_torchaudio("1.9.0")
    '0.9.0'
    >>> _determine_torchaudio("2.4.1")
    '2.4.1'
    >>> _determine_torchaudio("1.8.2")
    '0.9.1'

    """
    _version_exceptions = {
        "2.0.1": "2.0.2",
        "2.0.0": "2.0.1",
        "1.8.2": "0.9.1",
    }
    if torch_version in _version_exceptions:
        return _version_exceptions[torch_version]
    ver_major, ver_minor, ver_bugfix = map(int, torch_version.split("."))
    ta_ver_array = [ver_major, ver_minor, ver_bugfix]
    if ver_major == 1:
        ta_ver_array[0] = 0
    ta_ver_array[2] = ver_bugfix
    return ".".join(map(str, ta_ver_array))


def _determine_torchtext(torch_version: str) -> str:
    """Determine the torchtext version based on the torch version.

    >>> _determine_torchtext("1.9.0")
    '0.10.0'
    >>> _determine_torchtext("2.4.1")
    '0.18.0'
    >>> _determine_torchtext("1.8.2")
    '0.9.1'

    """
    _version_exceptions = {
        "2.0.1": "0.15.2",
        "2.0.0": "0.15.1",
        "1.8.2": "0.9.1",
    }
    if torch_version in _version_exceptions:
        return _version_exceptions[torch_version]
    ver_major, ver_minor, ver_bugfix = map(int, torch_version.split("."))
    tt_ver_array = [0, 0, 0]
    if ver_major == 1:
        tt_ver_array[1] = ver_minor + 1
        tt_ver_array[2] = ver_bugfix
    elif ver_major == 2:
        if ver_minor >= 3:
            tt_ver_array[1] = 18
        else:
            tt_ver_array[1] = ver_minor + 15
            tt_ver_array[2] = ver_bugfix
    else:
        raise ValueError(f"Invalid torch version: {torch_version}")
    return ".".join(map(str, tt_ver_array))


def _determine_torchvision(torch_version: str) -> str:
    """Determine the torchvision version based on the torch version.

    >>> _determine_torchvision("1.9.0")
    '0.10.0'
    >>> _determine_torchvision("2.4.1")
    '0.19.1'
    >>> _determine_torchvision("2.0.1")
    '0.15.2'

    """
    _version_exceptions = {
        "2.0.1": "0.15.2",
        "2.0.0": "0.15.1",
        "1.10.2": "0.11.3",
        "1.10.1": "0.11.2",
        "1.10.0": "0.11.1",
        "1.8.2": "0.9.1",
    }
    if torch_version in _version_exceptions:
        return _version_exceptions[torch_version]
    ver_major, ver_minor, ver_bugfix = map(int, torch_version.split("."))
    tv_ver_array = [0, 0, 0]
    if ver_major == 1:
        tv_ver_array[1] = ver_minor + 1
    elif ver_major == 2:
        tv_ver_array[1] = ver_minor + 15
    else:
        raise ValueError(f"Invalid torch version: {torch_version}")
    tv_ver_array[2] = ver_bugfix
    return ".".join(map(str, tv_ver_array))


def find_latest(ver: str) -> dict[str, str]:
    """Find the latest version.

    >>> from pprint import pprint
    >>> pprint(find_latest("2.4.1"))
    {'torch': '2.4.1',
     'torchaudio': '2.4.1',
     'torchtext': '0.18.0',
     'torchvision': '0.19.1'}
    >>> pprint(find_latest("2.1"))
    {'torch': '2.1.0',
     'torchaudio': '2.1.0',
     'torchtext': '0.16.0',
     'torchvision': '0.16.0'}

    """
    # drop all except semantic version
    ver = re.search(r"([\.\d]+)", ver).groups()[0]
    # in case there remaining dot at the end - e.g "1.9.0.dev20210504"
    ver = ver[:-1] if ver[-1] == "." else ver
    if not re.match(r"\d+\.\d+\.\d+", ver):
        ver += ".0"  # add missing bugfix
    logging.debug(f"finding ecosystem versions for: {ver}")

    # find first match
    return {
        "torch": ver,
        "torchvision": _determine_torchvision(ver),
        "torchtext": _determine_torchtext(ver),
        "torchaudio": _determine_torchaudio(ver),
    }


def adjust(requires: list[str], pytorch_version: Optional[str] = None) -> list[str]:
    """Adjust the versions to be paired within pytorch ecosystem.

    >>> from pprint import pprint
    >>> pprint(adjust([
    ...     "torch>=1.9.0", "torchvision>=0.10.0", "torchtext>=0.10.0", "torchaudio>=0.9.0", "vmaf-torch >=1.1.0"
    ... ], "2.1.0"))
    ['torch==2.1.0',
     'torchvision==0.16.0',
     'torchtext==0.16.0',
     'torchaudio==2.1.0',
     'vmaf-torch >=1.1.0']

    """
    if not pytorch_version:
        import torch

        pytorch_version = torch.__version__
    if not pytorch_version:
        raise ValueError(f"invalid torch: {pytorch_version}")

    requires_ = []
    options = find_latest(pytorch_version)
    logging.debug(f"determined ecosystem alignment: {options}")
    for req in requires:
        req_split = req.strip().split("#", maxsplit=1)
        # anything before fst # shall be requirements
        req = req_split[0].strip()
        # anything after # in the line is comment
        comment = "" if len(req_split) < 2 else "  #" + req_split[1]
        if not req:
            # if only comment, so make it short
            requires_.append(comment.strip())
            continue
        for lib, version in options.items():
            replace = f"{lib}=={version}" if version else ""
            req = re.sub(rf"^{lib}(?![-_\w]).*", replace, req)
        requires_.append(req + comment.rstrip())

    return requires_


def _offset_print(reqs: list[str], offset: str = "\t|\t") -> str:
    """Adding offset to each line for the printing requirements."""
    reqs = [offset + r for r in reqs]
    return os.linesep.join(reqs)


def main(requirements_path: str, torch_version: Optional[str] = None) -> None:
    """The main entry point with mapping to the CLI for positional arguments only."""
    # rU - universal line ending - https://stackoverflow.com/a/2717154/4521646
    with open(requirements_path, encoding="utf8") as fopen:
        requirements = fopen.readlines()
    requirements = adjust(requirements, torch_version)
    logging.info(
        f"requirements_path='{requirements_path}' with arg torch_version='{torch_version}' >>\n"
        f"{_offset_print(requirements)}"
    )
    with open(requirements_path, "w", encoding="utf8") as fopen:
        fopen.writelines([r + os.linesep for r in requirements])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    try:
        from jsonargparse import auto_cli, set_parsing_settings

        set_parsing_settings(parse_optionals_as_positionals=True)
        auto_cli(main, as_positional=False)
    except (ModuleNotFoundError, ImportError):
        logging.warning(
            "Expected `jsonargparse` is not installed,"
            " using the native `sys.argv` parser with positional arguments only."
            " Please install `pip install lightning_utilities[cli]` for better CLI experience."
        )
        cli_args = sys.argv[1:]
        if len(cli_args) < 1:
            logging.error("Expected at least 1 positional argument: <requirements_path> [<torch_version>]")
            sys.exit(1)
        # Use sys.argv directly if jsonargparse is not available
        main(*cli_args)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import warnings
from pathlib import Path

import litserve as ls

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.propagate = False
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

# COLOR CODES
RESET = "\u001b[0m"
RED = "\u001b[31m"
GREEN = "\u001b[32m"
BLUE = "\u001b[34m"
MAGENTA = "\u001b[35m"
BG_MAGENTA = "\u001b[45m"

# ACTION CODES
BOLD = "\u001b[1m"
UNDERLINE = "\u001b[4m"
INFO = f"{BOLD}{BLUE}[INFO]"
WARNING = f"{BOLD}{RED}[WARNING]"


def color(text, color_code, action_code=None):
    if action_code:
        return f"{action_code} {color_code}{text}{RESET}"
    return f"{color_code}{text}{RESET}"


REQUIREMENTS_FILE = "requirements.txt"
DOCKERFILE_TEMPLATE = """ARG PYTHON_VERSION=3.12
FROM python:$PYTHON_VERSION-slim

####### Add your own installation commands here #######
# RUN pip install some-package
# RUN wget https://path/to/some/data/or/weights
# RUN apt-get update && apt-get install -y <package-name>

WORKDIR /app
COPY . /app

# Install litserve and requirements
RUN pip install --no-cache-dir litserve=={version} {requirements}
EXPOSE {port}
CMD ["python", "/app/{server_filename}"]
"""

CUDA_DOCKER_TEMPLATE = """# Change CUDA and cuDNN version here
FROM nvidia/cuda:12.4.1-base-ubuntu22.04
ARG PYTHON_VERSION=3.12

ENV DEBIAN_FRONTEND=noninteractive
RUN apt-get update && apt-get install -y --no-install-recommends \\
        software-properties-common \\
        wget \\
    && add-apt-repository ppa:deadsnakes/ppa \\
    && apt-get update && apt-get install -y --no-install-recommends \\
        python$PYTHON_VERSION \\
        python$PYTHON_VERSION-dev \\
        python$PYTHON_VERSION-venv \\
    && wget https://bootstrap.pypa.io/get-pip.py -O get-pip.py \\
    && python$PYTHON_VERSION get-pip.py \\
    && rm get-pip.py \\
    && ln -sf /usr/bin/python$PYTHON_VERSION /usr/bin/python \\
    && ln -sf /usr/local/bin/pip$PYTHON_VERSION /usr/local/bin/pip \\
    && python --version \\
    && pip --version \\
    && apt-get purge -y --auto-remove software-properties-common \\
    && apt-get clean \\
    && rm -rf /var/lib/apt/lists/*

####### Add your own installation commands here #######
# RUN pip install some-package
# RUN wget https://path/to/some/data/or/weights
# RUN apt-get update && apt-get install -y <package-name>

WORKDIR /app
COPY . /app

# Install litserve and requirements
RUN pip install --no-cache-dir litserve=={version} {requirements}
EXPOSE {port}
CMD ["python", "/app/{server_filename}"]
"""

# Link our documentation as the bottom of this msg
SUCCESS_MSG = """{BOLD}{MAGENTA}Dockerfile created successfully{RESET}
Update {UNDERLINE}{dockerfile_path}{RESET} to add any additional dependencies or commands.{RESET}

{BOLD}Build the container with:{RESET}
> {UNDERLINE}docker build -t litserve-model .{RESET}

{BOLD}To run the Docker container on the machine:{RESET}
> {UNDERLINE}{RUN_CMD}{RESET}

{BOLD}To push the container to a registry:{RESET}
> {UNDERLINE}docker push litserve-model{RESET}
"""


def dockerize(server_filename: str, port: int = 8000, gpu: bool = False):
    """Generate a Dockerfile for the given server code.

    Example usage:
        litserve dockerize server.py --port 8000 --gpu

    Args:
        server_filename (str): The path to the server file. Example sever.py or app.py.
        port (int, optional): The port to expose in the Docker container.
        gpu (bool, optional): Whether to use a GPU-enabled Docker image.

    """
    requirements = ""
    if os.path.exists(REQUIREMENTS_FILE):
        requirements = f"-r {REQUIREMENTS_FILE}"
    else:
        warnings.warn(
            f"requirements.txt not found at {os.getcwd()}. "
            f"Make sure to install the required packages in the Dockerfile.",
            UserWarning,
        )

    current_dir = Path.cwd()
    if not (current_dir / server_filename).is_file():
        raise FileNotFoundError(f"Server file `{server_filename}` must be in the current directory: {os.getcwd()}")

    version = ls.__version__
    if gpu:
        run_cmd = f"docker run --gpus all -p {port}:{port} litserve-model:latest"
        docker_template = CUDA_DOCKER_TEMPLATE
    else:
        run_cmd = f"docker run -p {port}:{port} litserve-model:latest"
        docker_template = DOCKERFILE_TEMPLATE
    dockerfile_content = docker_template.format(
        server_filename=server_filename,
        port=port,
        version=version,
        requirements=requirements,
    )
    with open("Dockerfile", "w") as f:
        f.write(dockerfile_content)
    success_msg = SUCCESS_MSG.format(
        dockerfile_path=os.path.abspath("Dockerfile"),
        RUN_CMD=run_cmd,
        BOLD=BOLD,
        MAGENTA=MAGENTA,
        GREEN=GREEN,
        BLUE=BLUE,
        UNDERLINE=UNDERLINE,
        BG_MAGENTA=BG_MAGENTA,
        RESET=RESET,
    )
    print(success_msg)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import random
import time
from collections.abc import Generator

import psutil
import pytest
from fastapi import Request, Response
from fastapi.testclient import TestClient

from litserve.api import LitAPI
from litserve.server import LitServer
from litserve.utils import wrap_litserve_start


def pytest_collection_modifyitems(config, items):
    """Auto-mark tests based on their file path."""
    for item in items:
        # Get the relative path of the test file
        test_path = item.fspath.relto(config.rootdir)

        if "tests/unit/" in str(test_path):
            item.add_marker(pytest.mark.unit)
        elif "tests/integration/" in str(test_path):
            item.add_marker(pytest.mark.integration)
        elif "tests/e2e/" in str(test_path):
            item.add_marker(pytest.mark.e2e)


class SimpleLitAPI(LitAPI):
    def setup(self, device):
        self.model = lambda x: x**2

    def decode_request(self, request: Request):
        return request["input"]

    def predict(self, x):
        return self.model(x)

    def encode_response(self, output) -> Response:
        return {"output": output}


class SimpleStreamAPI(LitAPI):
    def setup(self, device) -> None:
        self.sentence = "LitServe is streaming output"

    def decode_request(self, request: Request) -> str:
        return request["prompt"]

    def predict(self, x) -> Generator:
        output = f"prompt={x} generated_output={self.sentence}".split()
        yield from output

    def encode_response(self, output: Generator) -> Generator:
        delay = 0.01  # delay for testing timeouts
        for out in output:
            time.sleep(delay)
            yield out.lower()


class SimpleBatchedStreamAPI(LitAPI):
    def setup(self, device) -> None:
        self.sentence = "LitServe is streaming output"

    def decode_request(self, request: Request) -> str:
        return request["prompt"]

    def batch(self, inputs):
        return inputs

    def predict(self, x) -> Generator:
        n = len(x)
        output = self.sentence.split()
        responses = [x]
        for out in output:
            responses.append([out] * n)
        yield from responses

    def encode_response(self, output: Generator) -> Generator:
        delay = 0.01  # delay for testing timeouts
        for out in output:
            time.sleep(delay)
            yield [e.lower() for e in out]

    def unbatch(self, output):
        yield from output


@pytest.fixture
def simple_litapi():
    return SimpleLitAPI()


@pytest.fixture
def simple_litapi_cls():
    return SimpleLitAPI


@pytest.fixture
def simple_stream_cls():
    return SimpleStreamAPI


@pytest.fixture
def simple_stream_api():
    return SimpleStreamAPI(stream=True)


@pytest.fixture
def simple_batched_stream_api():
    return SimpleBatchedStreamAPI()


@pytest.fixture
def lit_server(simple_litapi):
    server = LitServer(simple_litapi, accelerator="cpu", devices=1, timeout=10)
    with wrap_litserve_start(server) as s:
        yield s


@pytest.fixture
def sync_testclient(lit_server):
    with TestClient(lit_server.app) as client:
        yield client


@pytest.fixture
def killall():
    def _run(process):
        parent = psutil.Process(process.pid)
        for child in parent.children(recursive=True):
            child.kill()
        process.kill()

    return _run


@pytest.fixture
def openai_request_data():
    return {
        "model": "",
        "messages": [{"role": "string", "content": "string"}],
        "temperature": 0.7,
        "top_p": 1,
        "n": 1,
        "max_completion_tokens": 0,
        "stop": "string",
        "stream": False,
        "presence_penalty": 0,
        "frequency_penalty": 0,
        "user": "string",
    }


@pytest.fixture
def openai_response_data():
    return {
        "id": "chatcmpl-9dEtoQu4g45g3431SZ2s98S",
        "choices": [
            {
                "finish_reason": "stop",
                "index": 0,
                "logprobs": None,
                "message": {
                    "content": "10 + 6 is equal to 16.",
                    "role": "assistant",
                    "function_call": None,
                    "tool_calls": None,
                },
            }
        ],
        "created": 1719139092,
        "model": "gpt-3.5-turbo-0125",
        "object": "chat.completion",
        "system_fingerprint": None,
        "usage": {"completion_tokens": 10, "prompt_tokens": 25, "total_tokens": 35},
    }


@pytest.fixture
def openai_request_data_with_image():
    return {
        "model": "lit",
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "What's in this image?"},
                    {
                        "type": "image_url",
                        "image_url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                    },
                ],
            }
        ],
        "temperature": 0.7,
        "top_p": 1,
        "n": 1,
        "max_completion_tokens": 0,
        "stop": "string",
        "stream": False,
        "presence_penalty": 0,
        "frequency_penalty": 0,
        "user": "string",
    }


@pytest.fixture
def openai_request_data_with_audio_wav(openai_request_data):
    # Create a base64 encoded string from a list of audio data
    audio_data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    encoded_string = base64.b64encode(bytearray(audio_data)).decode("utf-8")

    request_data = openai_request_data.copy()
    request_data["messages"] = [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "What is in this recording?"},
                {"type": "input_audio", "input_audio": {"data": encoded_string, "format": "wav"}},
            ],
        },
    ]
    return request_data


@pytest.fixture
def openai_request_data_with_audio_flac(openai_request_data):
    # Create a base64 encoded string from a list of audio data
    audio_data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    encoded_string = base64.b64encode(bytearray(audio_data)).decode("utf-8")

    request_data = openai_request_data.copy()
    request_data["messages"] = [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "What is in this recording?"},
                {"type": "input_audio", "input_audio": {"data": encoded_string, "format": "flac"}},
            ],
        },
    ]
    return request_data


@pytest.fixture
def openai_request_data_with_tools():
    return {
        "model": "lit",
        "messages": [{"role": "user", "content": "What's the weather like in Boston today?"}],
        "tools": [
            {
                "type": "function",
                "function": {
                    "name": "get_current_weather",
                    "description": "Get the current weather in a given location",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                        },
                        "required": ["location"],
                    },
                },
            }
        ],
        "temperature": 0.7,
        "top_p": 1,
        "n": 1,
        "max_completion_tokens": 0,
        "stop": "string",
        "stream": False,
        "presence_penalty": 0,
        "frequency_penalty": 0,
        "user": "string",
    }


@pytest.fixture
def openai_request_data_with_response_format():
    return {
        "model": "lit",
        "messages": [
            {
                "role": "system",
                "content": "Extract the event information.",
            },
            {"role": "user", "content": "Alice and Bob are going to a science fair on Friday."},
        ],
        "response_format": {
            "type": "json_schema",
            "json_schema": {
                "name": "calendar_event",
                "schema": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "date": {"type": "string"},
                        "participants": {"type": "array", "items": {"type": "string"}},
                    },
                    "required": ["name", "date", "participants"],
                    "additionalProperties": "false",
                },
                "strict": "true",
            },
        },
        "temperature": 0.7,
        "top_p": 1,
        "n": 1,
        "max_completion_tokens": 0,
        "stop": "string",
        "stream": False,
        "presence_penalty": 0,
        "frequency_penalty": 0,
        "user": "string",
    }


@pytest.fixture
def openai_request_data_with_metadata():
    return {
        "model": "lit",
        "messages": [
            {
                "role": "system",
                "content": "Extract the event information.",
            },
            {"role": "user", "content": "Alice and Bob are going to a science fair on Friday."},
        ],
        "temperature": 0.7,
        "top_p": 1,
        "n": 1,
        "max_completion_tokens": 0,
        "stop": "string",
        "stream": False,
        "presence_penalty": 0,
        "frequency_penalty": 0,
        "user": "string",
        "metadata": {"user_id": "123", "trace_id": "abc-xyz"},
    }


@pytest.fixture
def openai_embedding_request_data():
    return {"input": "A beautiful sunset over the beach.", "model": "lit", "encoding_format": "float"}


@pytest.fixture
def openai_embedding_request_data_array():
    return {"input": ["A beautiful sunset over the beach."] * 4, "model": "lit", "encoding_format": "float"}


class MockEvent:
    def set(self):
        pass

    def wait(self):
        pass


class MockQueue:
    def put(self, item):
        pass

    def get(self, block=True, timeout=None):
        pass


@pytest.fixture
def mock_manager():
    class MockManager:
        def __init__(self):
            self.Queue = MockQueue
            self.Event = MockEvent
            self.dict = dict

        def shutdown(self):
            pass

    return MockManager()


@pytest.fixture
def port():
    return random.randint(10000, 65535)



import base64
import io
import os
from concurrent.futures import ThreadPoolExecutor

import PIL
import torch
import torchvision

import litserve as ls

device = "cuda" if torch.cuda.is_available() else "cpu"
device = "mps" if torch.backends.mps.is_available() else device
conf = {
    "cuda": {"batch_size": 16, "workers_per_device": 2},
    "cpu": {"batch_size": 8, "workers_per_device": 2},
    "mps": {"batch_size": 8, "workers_per_device": 2},
}

# Set float32 matrix multiplication precision if GPU is available and capable
if torch.cuda.is_available() and torch.cuda.get_device_capability() >= (8, 0):
    torch.set_float32_matmul_precision("high")


class ImageClassifierAPI(ls.LitAPI):
    def setup(self, device):
        print(device)
        weights = torchvision.models.ResNet18_Weights.DEFAULT
        self.image_processing = weights.transforms()
        self.model = torchvision.models.resnet18(weights=None).eval().to(device)
        self.pool = ThreadPoolExecutor(os.cpu_count())

    def decode_request(self, request):
        return request["image_data"]

    def batch(self, image_data_list):
        def process_image(image_data):
            image = base64.b64decode(image_data)
            pil_image = PIL.Image.open(io.BytesIO(image)).convert("RGB")
            return self.image_processing(pil_image)

        inputs = list(self.pool.map(process_image, image_data_list))
        return torch.stack(inputs).to(self.device)

    def predict(self, x):
        with torch.inference_mode():
            outputs = self.model(x)
            _, predictions = torch.max(outputs, 1)
        return predictions

    def unbatch(self, outputs):
        return outputs.tolist()

    def encode_response(self, output):
        return {"output": output}


def main(batch_size: int, workers_per_device: int):
    print(locals())
    api = ImageClassifierAPI(max_batch_size=batch_size, batch_timeout=0.01)
    server = ls.LitServer(
        api,
        timeout=10,
        workers_per_device=workers_per_device,
        fast_queue=True,
    )
    server.run(port=8000, log_level="warning")


if __name__ == "__main__":
    main(**conf[device])



#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config
import glob
import os
import re

import pt_lightning_sphinx_theme

import lightning_utilities
from lightning_utilities.docs import adjust_linked_external_docs, fetch_external_assets

# This function is used to populate the (source) links in the API
from lightning_utilities.docs.formatting import _linkcode_resolve

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.


_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_PATH_ROOT = os.path.realpath(os.path.join(_PATH_HERE, "..", ".."))
SPHINX_MOCK_REQUIREMENTS = int(os.environ.get("SPHINX_MOCK_REQUIREMENTS", True))

# -- Project information -----------------------------------------------------

# this name shall match the project name in Github as it is used for linking to code
project = "Lightning-Dev-Tools"
copyright = lightning_utilities.__copyright__
author = lightning_utilities.__author__

# The short X.Y version
version = lightning_utilities.__version__
# The full version, including alpha/beta/rc tags
release = lightning_utilities.__version__

# Options for the linkcode extension
# ----------------------------------
github_user = "Lightning-AI"
github_repo = project

# -- Project documents -------------------------------------------------------

fetch_external_assets(docs_folder=_PATH_HERE)

adjust_linked_external_docs("https://numpy.org/doc/stable/", "https://numpy.org/doc/{numpy.__version__}/", _PATH_HERE)
adjust_linked_external_docs(
    "https://pytorch.org/docs/stable/", "https://pytorch.org/docs/{torch.__version__}/", _PATH_HERE
)


# export the READme
def _convert_markdown(path_in: str, path_out: str, path_root: str) -> None:
    with open(path_in) as fp:
        readme = fp.read()
    # TODO: temp fix removing SVG badges and GIF, because PDF cannot show them
    readme = re.sub(r"(\[!\[.*\))", "", readme)
    readme = re.sub(r"(!\[.*.gif\))", "", readme)
    folder_names = (os.path.basename(p) for p in glob.glob(os.path.join(path_root, "*")) if os.path.isdir(p))
    for dir_name in folder_names:
        readme = readme.replace(f"]({dir_name}/", f"]({os.path.join(path_root, dir_name)}/")
    with open(path_out, "w") as fp:
        fp.write(readme)


_convert_markdown(os.path.join(_PATH_ROOT, "README.md"), "readme.md", _PATH_ROOT)

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.

needs_sphinx = "5.3"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    # 'sphinxcontrib.mockautodoc',  # raises error: directive 'automodule' is already registered ...
    # 'sphinxcontrib.fulltoc',  # breaks pytorch-theme with unexpected kw argument 'titles_only'
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.linkcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.imgmath",
    "myst_parser",
    "sphinx.ext.autosectionlabel",
    "nbsphinx",
    "sphinx_autodoc_typehints",
    "sphinx_paramlinks",
    "sphinx.ext.githubpages",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

myst_update_mathjax = False

# https://berkeley-stat159-f17.github.io/stat159-f17/lectures/14-sphinx..html#conf.py-(cont.)
# https://stackoverflow.com/questions/38526888/embed-ipython-notebook-in-sphinx-document
# I execute the notebooks manually in advance. If notebooks test the code,
# they should be run at build time.
nbsphinx_execute = "never"
nbsphinx_allow_errors = True
nbsphinx_requirejs_path = ""

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
# source_suffix = ['.rst', '.md', '.ipynb']
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
    ".ipynb": "nbsphinx",
}

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "PULL_REQUEST_TEMPLATE.md",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pt_lightning_sphinx_theme"
html_theme_path = [pt_lightning_sphinx_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme_options = {
    "pytorch_project": lightning_utilities.__homepage__,
    "canonical_url": lightning_utilities.__homepage__,
    "collapse_navigation": False,
    "display_version": True,
    "logo_only": False,
}

html_favicon = "_static/images/icon.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_templates", "_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = project + "-doc"

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
    # Latex figure (float) alignment
    "figure_align": "htbp",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, project + ".tex", project + " Documentation", author, "manual"),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, project, project + " Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        project,
        project + " Documentation",
        author,
        project,
        lightning_utilities.__docs__,
        "Miscellaneous",
    ),
]

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


def setup(app):
    # this is for hiding doctest decoration,
    # see: http://z4r.github.io/python/2011/12/02/hides-the-prompts-and-output/
    app.add_js_file("copybutton.js")


# Ignoring Third-party packages
# https://stackoverflow.com/questions/15889621/sphinx-how-to-exclude-imports-in-automodule
def _package_list_from_file(file: str) -> list[str]:
    list_pkgs = []
    with open(file) as fp:
        lines = fp.readlines()
    for ln in lines:
        found = [ln.index(ch) for ch in list(",=<>#") if ch in ln]
        pkg = ln[: min(found)] if found else ln
        if pkg.rstrip():
            list_pkgs.append(pkg.rstrip())
    return list_pkgs


# define mapping from PyPI names to python imports
PACKAGE_MAPPING = {
    "PyYAML": "yaml",
}
MOCK_PACKAGES = []
if SPHINX_MOCK_REQUIREMENTS:
    # mock also base packages when we are on RTD since we don't install them there
    MOCK_PACKAGES += _package_list_from_file(os.path.join(_PATH_ROOT, "requirements", "cli.txt"))
MOCK_PACKAGES = [PACKAGE_MAPPING.get(pkg, pkg) for pkg in MOCK_PACKAGES]

autodoc_mock_imports = MOCK_PACKAGES


# Resolve function
def linkcode_resolve(domain, info):
    return _linkcode_resolve(
        domain, info, github_user=github_user, github_repo=github_repo, main_branch="main", stable_branch="main"
    )


autosummary_generate = True

autodoc_member_order = "groupwise"
autoclass_content = "both"
# the options are fixed and will be soon in release,
#  see https://github.com/sphinx-doc/sphinx/issues/5459
autodoc_default_options = {
    "members": None,
    "methods": None,
    # 'attributes': None,
    "special-members": "__call__",
    "exclude-members": "_abc_impl",
    "show-inheritance": True,
    "private-members": True,
    "noindex": True,
}

# Sphinx will add “permalinks” for each heading and description environment as paragraph signs that
#  become visible when the mouse hovers over them.
# This value determines the text for the permalink; it defaults to "¶". Set it to None or the empty
#  string to disable permalinks.
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-html_add_permalinks
# html_add_permalinks = "¶"

# True to prefix each section label with the name of the document it is in, followed by a colon.
#  For example, index:Introduction for a section called Introduction that appears in document index.rst.
#  Useful for avoiding ambiguity when the same section heading appears in different documents.
# http://www.sphinx-doc.org/en/master/usage/extensions/autosectionlabel.html
autosectionlabel_prefix_document = True

# only run doctests marked with a ".. doctest::" directive
doctest_test_doctest_blocks = ""
doctest_global_setup = """
import os
"""
coverage_skip_undoc_in_source = True
